/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.data;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import javax.imageio.ImageIO;

import org.disperanto.data.IImageOperation;
import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.data.TdpClearRectangle;
import org.disperanto.data.TdpCopyImage;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpExternalLighting;
import org.disperanto.data.TdpInitializeWorkingMemory;
import org.disperanto.data.TdpLoadImage;
import org.disperanto.data.TdpManipulateMemorySlot;
import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpShowImage;
import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.data.TdpStoreImage;
import org.disperanto.data.TdpTimeoutSettings;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.CalculateCrcImagesCommand;
import org.disperanto.messages.ManipulateMemorySlotCommand;
import org.disperanto.messages.SetBrightnessCommand;
import org.disperanto.messages.SetExternalLightingCommand;
import org.disperanto.messages.SetTextCommand;
import org.disperanto.messages.SetTimeoutCommand;
import org.disperanto.messages.ShowImageCommand;
import org.disperanto.messages.StartSlideShowCommand;
import org.disperanto.messages.TdpCommandFactory;
import org.disperanto.messages.TdpMessage;
import org.disperanto.utils.TdpUtils;

/**
 * Handles the incoming messages
 */
public final class MessageHandler {

    private MessageHandler() {
        // Empty constructor
    }

    /**
     * Parses and handles the incoming commands and creates a list of responses.
     *
     * @param displays List of displays
     * @param commands List of incoming messages
     * @param logger Model of logging panel
     * @return List of responses
     */
    public static List<TdpMessage> handleCommands(List<Display> displays, List<TdpMessage> commands, ILogger logger) {
        final List<TdpMessage> responseList = new ArrayList<TdpMessage>();

        // Parse commands
        for (TdpMessage command : commands) {
            for (int address : command.getAddresses()) {
                Display display = findDisplay(displays, address);
                if (display == null) {
                    logger.log("Unknown address " + address + " received in command "
                            + command.getCommand().getName());
                    continue;
                }

                if (command.getMessageType() != ETdpMessageType.COMMAND) {
                    logger.log("Received non-command message: " + command.getMessageType());
                    responseList.add(TdpCommandFactory.createErrorNotification(address,
                            ETdpCommunicationError.ERROR_ILLEGAL_DATA));
                    continue;
                }

                try {
                    TdpMessage response = handleCommand(display, logger, command);
                    if (response == null) {
                        logger.log("Response was unexpectedly null for command: " + command);
                    } else {
                        responseList.add(response);
                    }
                } catch (TdpProtocolException e) {
                    responseList.add(TdpCommandFactory.createErrorNotification(address,
                            ETdpCommunicationError.ERROR_ILLEGAL_DATA));
                    continue;
                }

                // Add notifications
                if (display.getNotifications().size() > 0) {
                    responseList.add(createNotificationResponse(display, logger));
                }
            }
        }

        return responseList;
    }

    /**
     * @throws TdpProtocolException If an error occurred while trying to process the command. Unknown commands do no
     *             throw this error, but instead return an unknown command notification.
     */
    private static TdpMessage handleCommand(Display display, ILogger logger, TdpMessage command)
            throws TdpProtocolException {

        switch (command.getCommand()) {
        case DISPLAY_PROPERTIES:
            logger.log("DisplayProperties command received: " + command.toString());
            return createDisplayPropertiesResponse(display, command, logger);
        case STATUS:
            logger.log("Status command received: " + command.toString());
            return createStatusResponse(display, command, logger);
        case REBOOT:
            logger.log("Reboot command received: " + command.toString());
            return createRebootResponse(display, command, logger);
        case KEEP_ALIVE:
            logger.log("KeepAlive command received: " + command.toString());
            return createKeepAliveResponse(display, command, logger);
        case SET_COMMUNICATION_TIMEOUT:
            logger.log("SetCommunicationTimeout command received: " + command.toString());
            handleSetCommunicationTimeout(display, command, logger);
            return createSetCommunicationTimeoutResponse(display, command, logger);
        case SET_BRIGHTNESS:
            logger.log("SetBrightnessTable command received: " + command.toString());
            handleSetBrightness(display, command, logger);
            return createSetBrightnessResponse(display, command, logger);
        case SET_EXTERNAL_LIGHTING:
            logger.log("SetExternalLighting command received: " + command.toString());
            handleSetExternalLighting(display, command, logger);
            return createSetExternalLightingResponse(display, command, logger);
        case DIAGNOSTICS:
            logger.log("Diagnostics command received: " + command.toString());
            return createDiagnosticsResponse(display, command, logger);
        case MANIPULATE_MEMORY_SLOT:
            logger.log("ManipulateMemorySlot command received: " + command.toString());
            short manCrc = handleManipulateMemorySlot(display, command);
            return createManipulateMemorySlotResponse(display, command, manCrc, logger);
        case CALCULATE_CRC_IMAGES:
            logger.log("CalculateCrc command received: " + command.toString());
            final List<Short> crcs = handleCalculateCrc(display, command, logger);
            return createCalculateCrcResponse(display, command, crcs, logger);
        case SHOW_NO_IMAGE:
            logger.log("ShowNoImage command received: " + command.toString());
            handleShowNoImage(display, command);
            return createShowNoImageResponse(display, command, logger);
        case SHOW_IMAGE:
            logger.log("ShowImage command received: " + command.toString());
            short showImgCrc = handleShowImage(display, command, logger);
            return createShowImageResponse(display, command, showImgCrc, logger);
        case START_SLIDE_SHOW:
            logger.log("StartSlideShow command received: " + command.toString());
            List<Short> crcList = handleStartSlideShow(display, command, logger);
            return createStartSlideShowResponse(display, command, crcList, logger);
        case SUPPLIER_SERVICE_MODE:
            logger.log("ServiceMode command received: " + command.toString());
            return createServiceModeResponse(display, command, logger);
        case SET_TEXT:
            logger.log("SetText command received: " + command.toString());
            handleSetText(display, command, logger);
            return createSetTextResponse(display, command, logger);
        case NOTIFICATION:
            logger.log("ClearNotifications command received: " + command.toString());
            handleClearNotifications(display);
            return createNotificationResponse(display, logger);
        default:
            logger.log("Unknown command received: " + command.getCommand());
            int address = display.getAddress();
            return TdpCommandFactory.createErrorNotification(address, ETdpCommunicationError.ERROR_UNKNOWN_COMMAND);
        }
    }

    private static Display findDisplay(Iterable<Display> displays, int address) {
        for (Display display : displays) {
            if (display.getAddress() == address) {
                return display;
            }
        }
        return null;
    }

    private static void handleSetCommunicationTimeout(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.SET_COMMUNICATION_TIMEOUT) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            final TdpTimeoutSettings settings = SetTimeoutCommand.parse(message);
            logger.log("New timeout settings: " + settings.toString());
            currentDisplay.setTimeoutSettings(settings);
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }
    }

    private static void handleSetBrightness(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.SET_BRIGHTNESS) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            final TdpBrightnessTable brightnessTable = SetBrightnessCommand.parse(message);
            currentDisplay.setBrightnessTable(brightnessTable);
            logger.log("New brightness table: " +
                    SetBrightnessCommand.parse(message).getBrightnessTable().toString());
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }
    }

    private static void handleSetExternalLighting(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.SET_EXTERNAL_LIGHTING) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            TdpExternalLighting externalLighting = SetExternalLightingCommand.parse(message);
            currentDisplay.setExternalLighting(externalLighting);
            logger.log("External lighting set to: " + externalLighting.getState().getDescription());
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }
    }

    private static List<Short> handleCalculateCrc(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.CALCULATE_CRC_IMAGES) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            final List<Integer> imageSlots = CalculateCrcImagesCommand.parse(message);
            final List<Short> crcList = new ArrayList<Short>();

            for (Integer slot : imageSlots) {
                final BufferedImage image = currentDisplay.getBanks().getImage(slot);
                crcList.add(TdpUtils.calculateCrc(image));
            }

            return crcList;
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }

        return null;
    }

    private static short handleManipulateMemorySlot(Display currentDisplay, TdpMessage message)
            throws TdpProtocolException {
        if (message.getCommand() != ETdpCommand.MANIPULATE_MEMORY_SLOT) {
            throw new IllegalArgumentException("Wrong command id in message");
        }
        TdpManipulateMemorySlot data = ManipulateMemorySlotCommand.parse(message);

        TdpDisplayProperties displayProperties = currentDisplay.getDisplayProperties();

        BufferedImage canvas = null;
        for (IImageOperation operation : data.getOperations()) {
            switch (operation.getType()) {
            case INITIALIZE_WORKING_MEMORY:
                TdpInitializeWorkingMemory init = (TdpInitializeWorkingMemory)operation;
                canvas = new BufferedImage(init.getWidth(), init.getHeight(), BufferedImage.TYPE_INT_RGB);
                break;

            case CLEAR_RECTANGLE:
                TdpClearRectangle clear = (TdpClearRectangle)operation;
                Graphics g = canvas.getGraphics();
                g.setColor(Color.BLACK);
                g.fillRect(clear.getLeft(), clear.getTop(), clear.getWidth(), clear.getHeight());
                break;

            case LOAD_IMAGE:
                TdpLoadImage loadImg = (TdpLoadImage)operation;
                ByteArrayInputStream in = new ByteArrayInputStream(loadImg.getImage());
                try {
                    BufferedImage img = ImageIO.read(in);
                    canvas.getGraphics().drawImage(img, loadImg.getLeft(), loadImg.getTop(), null);
                } catch (IOException e) {
                    throw new TdpProtocolException("Incorrect image format", e);
                }
                break;

            case COPY_IMAGE:
                TdpCopyImage copyImg = (TdpCopyImage)operation;
                BufferedImage img = currentDisplay.getBanks().getImage(copyImg.getMemorySlot());
                canvas.getGraphics().drawImage(img, copyImg.getLeft(), copyImg.getTop(), null);
                break;

            case STORE_IMAGE:
                handleStoreImage(currentDisplay, displayProperties, canvas, operation);
                break;

            default:
                throw new IllegalStateException("Unknown type: " + operation.getType());
            }
        }

        return TdpUtils.calculateCrc(canvas);
    }

    private static void handleStoreImage(Display currentDisplay, TdpDisplayProperties displayProps,
            BufferedImage canvas, IImageOperation operation) throws TdpProtocolException {

        TdpStoreImage storeImg = (TdpStoreImage)operation;
        int storeSlot = storeImg.getMemorySlot();

        if (!displayProps.isWritableImage(storeSlot)) {
            throw new TdpProtocolException(
                    String.format(Locale.ROOT,
                            "Slot index is not modifiable: %d" + " (nrFixed=%d, nrWritable=%d)",
                            storeSlot, displayProps.getNrFixedImages(), displayProps.getNrWritableImages()));
        }

        currentDisplay.getBanks().setImage(storeSlot, canvas);

        ShowInfo showInfo = currentDisplay.getShowInfo();
        if (storeSlot == showInfo.getActiveImageSlot()) {
            // reset to update image
            showInfo.setActiveImageSlot(storeSlot);
        }
    }

    private static void handleShowNoImage(Display currentDisplay, TdpMessage message) {
        if (message.getCommand() != ETdpCommand.SHOW_NO_IMAGE) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        currentDisplay.getShowInfo().stopSlideShow();
        currentDisplay.getShowInfo().setNoImage();
    }

    private static short handleShowImage(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.SHOW_IMAGE) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            TdpShowImage imageInfo = ShowImageCommand.parse(message);
            int imageSlot = imageInfo.getImageSlot();

            currentDisplay.getShowInfo().stopSlideShow();
            currentDisplay.getShowInfo().setActiveImageSlot(imageSlot);

            TdpDisplayProperties displayProps = currentDisplay.getDisplayProperties();
            if (displayProps.isWritableImage(imageSlot)) {
                return TdpUtils.calculateCrc(currentDisplay.getBanks().getImage(imageSlot));
            } else if (displayProps.isFixedImage(imageSlot)) {
                // always return CRC zero for non-matrix slots
                return 0;
            } else {
                throw new TdpProtocolException(
                        String.format(Locale.ROOT, "Invalid slot index: %d" + " (nrFixed=%d, nrWritable=%d)", imageSlot,
                                displayProps.getNrFixedImages(), displayProps.getNrWritableImages()));
            }
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }

        return 0;
    }

    private static List<Short> handleStartSlideShow(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.START_SLIDE_SHOW) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            final TdpStartSlideShow slideShowInfo = StartSlideShowCommand.parse(message);
            final List<Short> crcList = new ArrayList<Short>();
            currentDisplay.getShowInfo().setSlideShow(slideShowInfo);

            for (Integer imageSlot : slideShowInfo.getImageSlotsSetting().keySet()) {
                crcList.add(TdpUtils.calculateCrc(currentDisplay.getBanks().getImage(imageSlot)));
            }

            return crcList;
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }

        return null;
    }

    private static void handleClearNotifications(Display currentDisplay) {
        currentDisplay.clearNotifications();
    }

    private static void handleSetText(Display currentDisplay, TdpMessage message, ILogger logger) {
        if (message.getCommand() != ETdpCommand.SET_TEXT) {
            throw new IllegalArgumentException("Wrong command id in message");
        }

        try {
            TdpSetText textLines = SetTextCommand.parse(message);
            currentDisplay.setTextLines(textLines);

            currentDisplay.getShowInfo().setText(textLines);
        } catch (TdpProtocolException e) {
            logger.log("TdpProtocol exception: " + e.getMessage());
        }
    }

    /**
     * Creates for the given display an response with display properties.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createDisplayPropertiesResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createDisplayPropertiesResponse(display.getAddress(),
                command.getMessageNumber(), display.getDisplayProperties());
        logger.log("DisplayProperties response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an response with status.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createStatusResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createStatusResponse(display.getAddress(),
                command.getMessageNumber(), display.getStatus());
        logger.log("Status response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an reboot response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createRebootResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createRebootResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("Reboot response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an keep-alive response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createKeepAliveResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createKeepAliveResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("KeepAlive response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an set communication timeout response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createSetCommunicationTimeoutResponse(Display display, TdpMessage command,
            ILogger logger) {

        final TdpMessage response = TdpCommandFactory.createSetCommunicationTimeoutResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("SetCommunicationTimeout response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an set brightness table response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createSetBrightnessResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createSetBrightnessTableResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("SetBrightnessTable response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an set external lighting response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createSetExternalLightingResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createSetExternalLightingResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("SetExternalLighting response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an diagnostic response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createDiagnosticsResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createDiagnosticsResponse(display.getAddress(),
                command.getMessageNumber(), display.getDiagnostics());
        logger.log("Diagnostics response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an manipulate memory slot response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createManipulateMemorySlotResponse(Display display, TdpMessage command, short crc,
            ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createManipulateMemorySlotResponse(display.getAddress(),
                command.getMessageNumber(), crc);
        logger.log("ManipulateMemorySlot response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an calculate crc response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createCalculateCrcResponse(Display display, TdpMessage command, List<Short> crcs,
            ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createCalculateCrcImagesResponse(display.getAddress(),
                command.getMessageNumber(), crcs);
        logger.log("CalculateCrc response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an ShowNoImage response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createShowNoImageResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createShowNoImageResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("ShowNoImage response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an ShowImage response.
     *
     * @param display Display the command is sent to
     * @param crc
     * @return TDP response
     */
    private static TdpMessage createShowImageResponse(Display display, TdpMessage command, short crc, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createShowImageResponse(display.getAddress(),
                command.getMessageNumber(), crc);
        logger.log("ShowImage response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an StartSlideShow response.
     *
     * @param display Display the command is sent to
     * @param crcList
     * @return TDP response
     */
    private static TdpMessage createStartSlideShowResponse(Display display, TdpMessage command, List<Short> crcList,
            ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createStartSlideShowResponse(display.getAddress(),
                command.getMessageNumber(), crcList);
        logger.log("StartSlideShow response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an ServiceMode response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createServiceModeResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createServiceModeResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("ServiceMode response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an SetText response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createSetTextResponse(Display display, TdpMessage command, ILogger logger) {
        final TdpMessage response = TdpCommandFactory.createSetTextResponse(display.getAddress(),
                command.getMessageNumber());
        logger.log("SetText response created: " + response.toString());

        return response;
    }

    /**
     * Creates for the given display an notification response.
     *
     * @param display Display the command is sent to
     * @return TDP response
     */
    private static TdpMessage createNotificationResponse(Display display, ILogger logger) {
        TdpMessage response = TdpCommandFactory.createNotification(display.getAddress(), display.getNotifications());
        logger.log("Notification response created: " + response.toString());

        return response;
    }

}
