/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.data;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.displaysimulator.listeners.IDisplayListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Information about current state of shown image
 */
public class ShowInfo {
    private static final Logger LOG = LoggerFactory.getLogger(ShowInfo.class);

    private SlideShowThread slideShow;

    private TdpBrightnessTable brightnessTable;

    /**
     * If true, the display is enabled.
     * If false, the display is powered down (e.g. by sending the "no image" command)
     */
    private boolean isDisplayEnabled = true;

    /**
     * The currently active image slot.
     * This is only used when the slide show is not active.
     */
    private int activeImageSlot = 0;

    private final List<IDisplayListener> listeners = new CopyOnWriteArrayList<IDisplayListener>();

    /**
     * The currently shown text
     */
    private TdpSetText shownText;

    /**
     * @return the slideShow
     */
    public final TdpStartSlideShow getSlideShow() {
        if (slideShow != null) {
            return slideShow.getSlideShowInfo();
        }
        return null;
    }

    public final void stopSlideShow() {
        if (slideShow != null) {
            slideShow.stop();
            slideShow.clearListeners();
        }
    }

    /**
     * @param slideShowSettings the slideShow to set
     */
    public final void setSlideShow(TdpStartSlideShow slideShowSettings) {
        stopSlideShow();
        if (slideShowSettings != null) {
            slideShow = new SlideShowThread(slideShowSettings);

            for (IDisplayListener listener : listeners) {
                slideShow.addListener(listener);
            }

            final ExecutorService threadExecutor = Executors.newFixedThreadPool(1);
            threadExecutor.execute(slideShow);
        }
    }

    /**
     * @return the brightnessTable
     */
    public final TdpBrightnessTable getBrightnessTable() {
        return brightnessTable;
    }

    /**
     * @param brightnessTable the brightnessTable to set
     */
    public final void setBrightnessTable(TdpBrightnessTable brightnessTable) {
        this.brightnessTable = brightnessTable;
    }

    /**
     * @return the activeImageSlot
     */
    public final int getActiveImageSlot() {
        return activeImageSlot;
    }

    /**
     * @param activeImageSlot the activeImage to set (set to null if slide show is active)
     */
    public final void setActiveImageSlot(int activeImageSlot) {
        this.isDisplayEnabled = true;
        this.activeImageSlot = activeImageSlot;
        fireImageChange();
    }

    /**
     * @param textLines the text for the text display
     */
    public final void setText(TdpSetText textLines) {
        this.isDisplayEnabled = true;
        this.shownText = textLines;
        fireTextChange();
    }

    /**
     * @return true if the display is enabled, false if not
     */
    public boolean isDisplayEnabled() {
        return isDisplayEnabled;
    }

    /**
     * Disables the display.
     */
    public void setNoImage() {
        this.isDisplayEnabled = false;
        fireImageChange();
    }

    /**
     * Adds a listener for image changes
     *
     * @param listener Listener for images changes
     */
    public final void addListener(IDisplayListener listener) {
        listeners.add(listener);
    }

    public final void removeListener(IDisplayListener listener) {
        listeners.remove(listener);
    }

    /**
     * Notifies a receive to the listeners.
     */
    private void fireImageChange() {
        LOG.debug("Fire image change for active image slot: " + activeImageSlot + ", display enabled: " +
                isDisplayEnabled);
        for (IDisplayListener listener : listeners) {
            listener.imageChanged(activeImageSlot, isDisplayEnabled);
        }
    }

    /**
     * Notifies a receive to the listeners.
     */
    private void fireTextChange() {
        LOG.debug("Fire text change, display enabled: " + isDisplayEnabled);
        for (IDisplayListener listener : listeners) {
            listener.textChanged(shownText, isDisplayEnabled);
        }
    }

}
