/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.data;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.displaysimulator.listeners.IDisplayListener;
import org.disperanto.displaysimulator.utils.ThreadHelper;
import org.disperanto.enums.ETdpSlideShowMode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Changes the slides of a slide show.
 */
public class SlideShowThread implements Runnable {
    private static final Logger LOG = LoggerFactory.getLogger(SlideShowThread.class);

    private final List<IDisplayListener> listeners = new ArrayList<IDisplayListener>();
    private volatile boolean active = true;
    private final TdpStartSlideShow slideShowInfo;

    /**
     * Constructor
     */
    public SlideShowThread(TdpStartSlideShow slideShowInfo) {
        this.slideShowInfo = slideShowInfo;
    }

    /**
     * @return the slideShowInfo
     */
    public final TdpStartSlideShow getSlideShowInfo() {
        return slideShowInfo;
    }

    @Override
    public final void run() {
        LOG.debug("Slide show started");

        while (active) {
            final Iterator<Integer> it = slideShowInfo.getImageSlotsSetting().keySet().iterator();
            while (active && it.hasNext()) {
                final Integer imageSlot = it.next();

                fireChange(imageSlot);
                ThreadHelper.sleep(slideShowInfo.getImageSlotsSetting().get(imageSlot) * 100); // timeout is given
                                                                                               // in tenth
            }

            if (slideShowInfo.getMode() == ETdpSlideShowMode.SHOW_ONCE) {
                // end slide show after last image is put on display.
                break;
            }
        }

        LOG.debug("Slide show stopped");
    }

    /**
     * Stops the thread
     */
    public final void stop() {
        active = false;
    }

    /**
     * Adds a listener for image changes.
     * 
     * @param listener Listener for images changes
     */
    public final void addListener(IDisplayListener listener) {
        listeners.add(listener);
    }

    /**
     * Adds a listener for image changes.
     */
    public final void clearListeners() {
        listeners.clear();
    }

    /**
     * Notifies a receive to the listeners.
     */
    private void fireChange(Integer imageSlot) {
        LOG.debug("Fire change for slide image slot: " + imageSlot);
        for (IDisplayListener listener : listeners) {
            listener.imageChanged(imageSlot, true);
        }
    }
}
