/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.ui.components;

import java.awt.Dimension;

import javax.swing.InputVerifier;
import javax.swing.JComponent;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Input field for a number.
 */
public class NumericTextField extends JTextField {
    private static final Logger LOG = LoggerFactory.getLogger(NumericTextField.class);

    /**
     * Default preferred width in pixels.
     */
    private static final int DEFAULT_PREFERRED_WIDTH = 50;

    private long minValue;
    private long maxValue;
    private final NumericDocument document;

    public NumericTextField() {
        this(0, Integer.MAX_VALUE);
    }

    public NumericTextField(long value) {
        this(0, Long.MAX_VALUE);
        setValue(value);
    }

    public NumericTextField(long minValue, long maxValue) {
        this.minValue = minValue;
        this.maxValue = maxValue;

        setText(Long.toString(minValue));

        document = new NumericDocument();
        document.setNegativeAccepted(minValue < 0);
        setDocument(document);

        setInputVerifier(new InputVerifier() {
            @Override
            public boolean verify(JComponent input) {
                return isValueValid();
            }
        });

        setHorizontalAlignment(SwingConstants.RIGHT);
        setPreferredSize(new Dimension(DEFAULT_PREFERRED_WIDTH, (int)getPreferredSize().getHeight()));
    }

    public NumericTextField(long value, long minValue, long maxValue) {
        this(minValue, maxValue);
        setValue(value);
    }

    public void setRange(long minValue, long maxValue) {
        this.minValue = minValue;
        this.maxValue = maxValue;

        document.setNegativeAccepted(minValue < 0);
    }

    public void setValue(long value) {
        if (value < minValue) {
            LOG.warn("Value below minimum (" + minValue + "): " + value);
        }
        if (value > maxValue) {
            LOG.warn("Value above maximum (" + maxValue + "): " + value);
        }
        setText(Long.toString(value));
    }

    public long getValue() {
        try {
            return Long.parseLong(getText());
        } catch (NullPointerException e) {
            return minValue;
        } catch (NumberFormatException e) {
            return minValue;
        }
    }

    public boolean containsValue() {
        try {
            Long.parseLong(getText());
            return true;
        } catch (NullPointerException e) {
            return false;
        } catch (NumberFormatException e) {
            return false;
        }
    }

    /**
     * Determines if the value as typed by the user is valid and within the minimum and maximum range.
     *
     * Note that getValue() always returns a valid long, but if the entered value is out of range
     * (e.g. larger than Integer.MAX_INT), then getValue() will simply return the minimum value
     *
     * @return true if the value is valid, false if not
     */
    public boolean isValueValid() {
        long value = 0;
        try {
            value = Long.parseLong(getText());
        } catch (Exception e) {
            return false;
        }
        if (value < minValue) {
            return false;
        }
        if (value > maxValue) {
            return false;
        }
        return true;
    }

    private static class NumericDocument extends PlainDocument {
        private static final String NUMBER_FORMAT = "[0-9+]*";
        private static final String NUMBER_FORMAT_ALLOW_NEGATIVE  = "[-]?[0-9+]*";

        private static final char PLUS = '+';
        private static final char MINUS = '-';

        private String matchFormat;
        private boolean negativeAccepted = false;

        public NumericDocument() {
            matchFormat = NUMBER_FORMAT;
        }

        public void setNegativeAccepted(boolean negativeaccepted) {
            negativeAccepted = negativeaccepted;
            matchFormat = negativeaccepted ? NUMBER_FORMAT_ALLOW_NEGATIVE : NUMBER_FORMAT;
        }

        @Override
        public void insertString(int offset, String str, AttributeSet attr) throws BadLocationException {
            if (str == null) {
                return;
            }

            if (!str.matches(matchFormat)) {
                return;
            }

            // Insertion of a minus sign (-) is only allowed at the left side of the number.
            // Typing + or - anywhere else changes the sign of the number to the requested type.
            if (negativeAccepted) {
                if (str.indexOf(PLUS) != -1) {
                    if (isNegative()) {
                        super.remove(0, 1);
                    }
                    return;
                } else if (str.indexOf(MINUS) != -1) {
                    if (str.indexOf(MINUS) != 0 || offset != 0) {
                        if (!isNegative()) {
                            super.insertString(0, "-", attr);
                        }
                        return;
                    }
                }
            }

            // Ensure that no numbers are typed before the minus sign (leading to "123-456", which would be invalid)
            if (isNegative() && (offset == 0)) {
                return;
            }

            // Never insert the + symbol
            if (str.indexOf(PLUS) != -1) {
                return;
            }

            super.insertString(offset, str, attr);
        }

        /**
         * Determines if the currently visible input starts with a minus symbol (-).
         */
        private boolean isNegative() throws BadLocationException {
            return getText(0, 1).indexOf(MINUS) != -1;
        }
    }

}
