/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.ui.controllers;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpSetText;
import org.disperanto.displaysimulator.data.Display;
import org.disperanto.displaysimulator.data.DisplayController;
import org.disperanto.displaysimulator.data.ShowInfo;
import org.disperanto.displaysimulator.listeners.IDisplayControllerListener;
import org.disperanto.displaysimulator.listeners.IDisplayListener;
import org.disperanto.displaysimulator.listeners.IPropertiesListener;
import org.disperanto.displaysimulator.ui.models.DisplayModel;
import org.disperanto.displaysimulator.ui.panels.ImagePanel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * View controller for ImageViewPanel.
 */
public class ImageViewController implements ActionListener, ChangeListener, IDisplayControllerListener,
        IDisplayListener, IPropertiesListener {
    private static final Logger LOG = LoggerFactory.getLogger(ImageViewController.class);

    private final DisplayModel model;
    private final ImagePanel panel;
    private final DisplayController controller;
    private int displayAddress;

    /**
     * Constructor
     *
     * @param panel Panel that is controlled
     * @param model Model of panel
     * @param controller Display
     */
    public ImageViewController(ImagePanel panel, DisplayModel model, DisplayController controller) {
        this.model = model;
        this.panel = panel;
        this.controller = controller;
    }

    private void initDisplay(int imageSlot, boolean isDisplayEnabled) {
        panel.getCurrentlyShownRadio().setSelected(true);
        model.setActiveImageSlot(imageSlot);
        model.setDisplayEnabled(isDisplayEnabled);
        model.setShownImage(controller.getDisplay(displayAddress).getBanks().getImage(imageSlot));
        model.setShownText(controller.getDisplay(displayAddress).getTextLines());
    }

    @Override
    public final void actionPerformed(ActionEvent a) {
        LOG.debug("Image bank changed");
        setImage();
    }

    @Override
    public final void stateChanged(ChangeEvent arg0) {
        LOG.debug("Spinner changed");
        setImage();
    }

    private void setImage() {
        int imageSlot = model.getActiveImageSlot();
        if (panel.getImageSlotRadio().isSelected()) {
            // use selected image slot from panel
            imageSlot = (Integer)panel.getImageSlot().getValue();
        }

        Display display = controller.getDisplay(displayAddress);
        if (display == null) {
            model.setShownImage(null);
            return;
        }

        TdpDisplayProperties props = display.getDisplayProperties();
        if (props.isFixedImage(imageSlot)) {
            model.setShownImage(createFixedImage(props, imageSlot));
        } else {
            model.setShownImage(display.getBanks().getImage(imageSlot));
        }
    }

    @Override
    public final void displayChanged(int activeDisplay) {
        if (activeDisplay != displayAddress) {
            displayAddress = activeDisplay;

            // switch to other display
            ShowInfo info = controller.getDisplay(displayAddress).getShowInfo();
            initDisplay(info.getActiveImageSlot(), info.isDisplayEnabled());
        }
    }

    @Override
    public final void imageChanged(int imageSlot, boolean isDisplayEnabled) {
        model.setActiveImageSlot(imageSlot);
        model.setDisplayEnabled(isDisplayEnabled);
        setImage();
    }

    private void setText(TdpSetText textLines) {
        Display display = controller.getDisplay(displayAddress);
        if (display == null) {
            model.setShownText(null);
            return;
        }
        model.setShownText(textLines);
    }

    @Override
    public void textChanged(TdpSetText text, boolean isDisplayEnabled) {
        model.setDisplayEnabled(isDisplayEnabled);
        setText(text);
    }

    @Override
    public void propertiesChanged(int displayAddress, TdpDisplayProperties properties) {
        // Number of fixed/writable slots may have changed, refresh image
        setImage();
    }

    private static BufferedImage createFixedImage(TdpDisplayProperties props, int imageSlot) {
        int w = props.getDisplayWidthInPixels();
        int h = props.getDisplayHeightInPixels();
        if (w == 0 || h == 0) {
            // Width/height may not exist for non-matrix displays
            w = 192;
            h = 180;
        }
        
        String text = "Fixed image #" + imageSlot;
        
        BufferedImage image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        Graphics2D g = image.createGraphics();
        g.setColor(Color.WHITE);
        
        Rectangle r = g.getFontMetrics().getStringBounds(text, g).getBounds();
        g.drawString(text, (w - r.width) / 2, (h + r.height) / 2);
        
        g.dispose();
        return image;
    }

}
