/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.ui.models;

import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpSetText.TdpTextLine;
import org.disperanto.displaysimulator.listeners.IImageListener;
import org.disperanto.displaysimulator.listeners.ITextListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Model for panel with display properties.
 */
public class DisplayModel {
    private static final Logger LOG = LoggerFactory.getLogger(DisplayModel.class);

    private int displayAddress;

    /**
     * If true the display is enabled. If false, the display is powered down.
     */
    private boolean isDisplayEnabled = true;

    /**
     * Image slot of active image on display.
     */
    private int activeImageSlot;

    private BufferedImage shownImage;

    private String shownText;

    private final List<IImageListener> imageListeners = new ArrayList<IImageListener>();
    private final List<ITextListener> textListeners = new ArrayList<ITextListener>();

    /**
     * @return the activeImageSlot
     */
    public final int getActiveImageSlot() {
        return activeImageSlot;
    }

    /**
     * @param activeImageSlot the activeImageSlot to set
     */
    public final void setActiveImageSlot(int activeImageSlot) {
        this.activeImageSlot = activeImageSlot;
    }

    /**
     * @return true if the display is enabled, false if not
     */
    public boolean isDisplayEnabled() {
        return isDisplayEnabled;
    }

    /**
     * @param isDisplayEnabled if true enables the display, if false powers it down
     */
    public void setDisplayEnabled(boolean isDisplayEnabled) {
        this.isDisplayEnabled = isDisplayEnabled;
        fireImageChange();
    }

    /**
     * @param shownImage the shownImage to set
     */
    public final void setShownImage(BufferedImage shownImage) {
        this.shownImage = shownImage;
        fireImageChange();
    }

    /**
     * @return the shownImage
     */
    public final BufferedImage getShownImage() {
        return shownImage;
    }

    /**
     * @param textLines the textLines to set
     */
    public void setShownText(TdpSetText textLines) {
        if (textLines == null) {
            this.shownText = "";
        } else {
            StringBuilder builder = new StringBuilder();
            for (TdpTextLine line : textLines.getLines()) {
                builder.append(line.getAlignment()).append(":\t").append(line.getText()).append("\n");
            }
            this.shownText = builder.toString();
        }
        fireTextChange();
    }

    /**
     * @return the shownText
     */
    public String getShownText() {
        return shownText;
    }

    /**
     * @param listener Listener of changes in image
     */
    public final void addListener(IImageListener listener) {
        imageListeners.add(listener);
    }

    /**
     * @param listener Listener of changes in text
     */
    public final void addListener(ITextListener listener) {
        textListeners.add(listener);
    }

    /**
     * Fires notification for listeners.
     */
    public final void fireImageChange() {
        LOG.debug("Fire image change");
        for (IImageListener listener : imageListeners) {
            listener.imageChanged(displayAddress, shownImage, isDisplayEnabled);
        }
    }

    /**
     * Fires notification for listeners.
     */
    public final void fireTextChange() {
        LOG.debug("Fire text change");
        for (ITextListener listener : textListeners) {
            listener.textChanged(shownText, isDisplayEnabled);
        }
    }

}
