package org.disperanto.displaysimulator.ui.panels;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.disperanto.displaysimulator.data.Display;
import org.disperanto.displaysimulator.data.DisplayController;
import org.disperanto.displaysimulator.data.DisplayCreator;
import org.disperanto.displaysimulator.listeners.IAvailableDisplayListener;
import org.disperanto.displaysimulator.listeners.IDisplayControllerListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.miginfocom.swing.MigLayout;

public class DisplayListPanel extends JPanel {

    private static final Logger LOG = LoggerFactory.getLogger(DisplayListPanel.class);

    private final List<IDisplayControllerListener> externalListeners =
            new CopyOnWriteArrayList<IDisplayControllerListener>();
    private final DisplayCreator displayCreator = new DisplayCreator();
    
    private final DisplayController controller;
    private final JList<Display> displayList;
    private final JButton removeButton;

    /** If greater than zero, ignore list selection events */
    private int suppressSelectionEvents;

    public DisplayListPanel(DisplayController controller) {
        this.controller = controller;

        JButton addButton = new JButton("Add display");
        addButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doAddDisplay();
            }
        });
        
        removeButton = new JButton("Remove display");
        removeButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                doRemoveDisplay();
            }
        });

        displayList = new JList<Display>();
        displayList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        displayList.addListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent e) {
                if (suppressSelectionEvents <= 0) {
                    updateButtons();
                    fireDisplayChange();
                }
            }
        });
        displayList.setCellRenderer(new DefaultListCellRenderer() {
            @Override
            public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected,
                    boolean cellHasFocus) {

                if (value instanceof Display) {
                    Display display = (Display)value;
                    value = String.format(Locale.ROOT, "Display [%02d]", display.getAddress());
                }
                return super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
            }
        });
        
        updateDisplayList();
        controller.addAvailableDisplayListener(new IAvailableDisplayListener() {
            @Override
            public void onDisplayAdded(Display display) {
                updateDisplayList();
            }

            @Override
            public void onDisplayRemoved(Display display) {
                updateDisplayList();
            }
        });

        setBorder(BorderFactory.createTitledBorder("Available displays"));
        setMinimumSize(new Dimension(200, 100));
        setLayout(new MigLayout("", "[][]", "[][][grow]"));
        add(new JScrollPane(displayList), "width 200, growy, spany");
        add(addButton, "grow, wrap");
        add(removeButton, "grow, wrap");
    }

    private void doAddDisplay() {
        String msg = "Please enter an address for the new display";
        String rawValue = JOptionPane.showInputDialog(this, msg, "1");
        if (rawValue == null) {
            return; // Canceled
        }

        int address = -1;
        try {
            address = Integer.parseInt(rawValue);
        } catch (NumberFormatException nfe) {
            // Invalid input
        }
        
        if (address >= 0) {
            Display display = displayCreator.createDisplay(address);
            controller.addDisplay(display);
            displayList.setSelectedValue(display, true);
        } else {
            JOptionPane.showMessageDialog(this, "Invalid address. Please enter a positive integer.");
        }
    }

    private void doRemoveDisplay() {
        controller.removeDisplay(getActiveDisplay());
    }

    private void updateDisplayList() {
        Display oldSelection = displayList.getSelectedValue();

        suppressSelectionEvents++;
        List<Display> displays = controller.getDisplays();
        displayList.setListData(displays.toArray(new Display[0]));
        suppressSelectionEvents--;

        if (oldSelection != null && displays.contains(oldSelection)) {
            // Restore old selection
            displayList.setSelectedValue(oldSelection, true);
        } else if (!displays.isEmpty()) {
            // Set initial selection
            displayList.setSelectedIndex(0);
        }

        updateButtons();
    }

    private void updateButtons() {
        // Ensure a display is selected and disallow removing the final display (causes problems for some panels)
        removeButton.setEnabled(controller.getDisplays().size() > 1 && getActiveDisplay() >= 0);
    }

    private int getActiveDisplay() {
        Display selected = displayList.getSelectedValue();
        return (selected != null ? selected.getAddress() : -1);
    }

    public void addSelectionListener(IDisplayControllerListener ls) {
        externalListeners.add(ls);
    }

    /**
     * fires notification for listeners.
     */
    public final void fireDisplayChange() {
        LOG.debug("Fire change for display");

        int activeDisplay = getActiveDisplay();
        if (activeDisplay < 0) {
            LOG.debug("No display is active");
            return;
        }

        for (IDisplayControllerListener ls : externalListeners) {
            ls.displayChanged(activeDisplay);
        }
    }

}
