/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
   list of conditions and the following disclaimer in the documentation and/or
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 */
package org.disperanto.displaysimulator.ui.panels;

import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.displaysimulator.listeners.IPropertiesListener;
import org.disperanto.displaysimulator.ui.components.NumericTextField;
import org.disperanto.enums.ETdpDisplayType;
import org.disperanto.messages.SetTextCommand;

import net.miginfocom.swing.MigLayout;

/**
 * Panel for showing display properties.
 */
public class PropertiesPanel extends JPanel implements IPropertiesListener {

    // components
    private JLabel addressField;
    private NumericTextField protocolField;
    private JComboBox<ETdpDisplayType> displayTypeField;
    private JTextField supplierAndProductField;
    private JTextField serialField;
    private JTextField softwareVersionField;
    private JCheckBox externalLightingCheck;
    private NumericTextField displayHeightField;
    private NumericTextField displayWidthField;
    private NumericTextField numberFixedImagesField;
    private NumericTextField numberWritableImagesField;
    private JCheckBox slideShowCheck;
    private JCheckBox rgbColorsCheck;
    private JTextField rgbColorsField;
    private JCheckBox fixedColorPaletteCheck;
    private JTextField fixedColorPaletteField;
    private JCheckBox pngSupportedCheck;
    private NumericTextField textDisplayRowsField;
    private NumericTextField textDisplayColumnsField;

    private JButton saveButton;

    /**
     * Constructor
     */
    public PropertiesPanel() {
        initUI();
    }

    /**
     * Add listener for panel actions
     *
     * @param listener Action listener
     */
    public final void addListener(ActionListener listener) {
        saveButton.addActionListener(listener);
    }

    private void initUI() {
        setLayout(new MigLayout());
        setBorder(BorderFactory.createTitledBorder("Properties"));

        addressField = new JLabel();
        protocolField = new NumericTextField(1, 255);
        displayTypeField = new JComboBox<ETdpDisplayType>(ETdpDisplayType.values());
        supplierAndProductField = new JTextField(10);
        serialField = new JTextField(10);
        softwareVersionField = new JTextField(10);
        externalLightingCheck = new JCheckBox();
        displayHeightField = new NumericTextField(1, 500);
        displayWidthField = new NumericTextField(1, 1000);
        numberFixedImagesField = new NumericTextField(0, 16383);
        numberWritableImagesField = new NumericTextField(0, 16383);
        slideShowCheck = new JCheckBox();
        rgbColorsCheck = new JCheckBox();
        rgbColorsField = new JTextField(10);
        fixedColorPaletteCheck = new JCheckBox();
        fixedColorPaletteField = new JTextField(10);
        pngSupportedCheck = new JCheckBox();
        textDisplayRowsField = new NumericTextField(1, SetTextCommand.MAX_ROWS);
        textDisplayColumnsField = new NumericTextField(1, SetTextCommand.MAX_CHARACTERS_ROW);

        add(new JLabel("Display address:"));
        add(addressField, "wrap, grow");

        add(new JLabel("protocol version:"));
        add(protocolField, "wrap, grow");

        add(new JLabel("Display type:"));
        add(displayTypeField, "wrap, grow");

        add(new JLabel("Supplier and product type:"));
        add(supplierAndProductField, "wrap, grow");

        add(new JLabel("Serial number:"));
        add(serialField, "wrap, grow");

        add(new JLabel("Software version:"));
        add(softwareVersionField, "wrap, grow");

        add(new JLabel("External lighting present:"));
        add(externalLightingCheck, "wrap");

        add(new JLabel("Display width:"));
        add(displayWidthField, "wrap, grow");

        add(new JLabel("Display height:"));
        add(displayHeightField, "wrap, grow");

        add(new JLabel("Number of fixed images:"));
        add(numberFixedImagesField, "wrap, grow");

        add(new JLabel("Number of writable images:"));
        add(numberWritableImagesField, "wrap, grow");

        add(new JLabel("Slide show possible:"));
        add(slideShowCheck, "wrap");

        add(new JLabel("RGB colors:"));
        add(rgbColorsCheck, "split 2");
        add(rgbColorsField, "wrap, grow");

        add(new JLabel("Fixed color palette:"));
        add(fixedColorPaletteCheck, "split 2");
        add(fixedColorPaletteField, "wrap, grow");

        JLabel pngSupportedLabel = new JLabel("PNG supported:*");
        pngSupportedLabel.setToolTipText("PNG support is mandatory from v2.0 onwards");
        add(pngSupportedLabel);
        add(pngSupportedCheck, "wrap");

        add(new JLabel("Text display height:"));
        add(textDisplayRowsField, "wrap, grow");

        add(new JLabel("Text display width:"));
        add(textDisplayColumnsField, "wrap, grow");

        add(new JLabel(""));
        saveButton = new JButton("Apply");
        add(saveButton, "wrap");
    }

    @Override
    public final void propertiesChanged(int displayAddress, TdpDisplayProperties properties) {
        addressField.setText("" + displayAddress);
        protocolField.setValue(properties.getProtocolVersion());
        displayTypeField.setSelectedItem(properties.getDisplayType());
        supplierAndProductField.setText(properties.getSupplierAndProductType());
        serialField.setText(properties.getSerialNumber());
        softwareVersionField.setText(properties.getSoftwareVersion());
        externalLightingCheck.setSelected(properties.hasExternalLighting());
        displayHeightField.setValue(properties.getDisplayHeightInPixels());
        displayWidthField.setValue(properties.getDisplayWidthInPixels());
        numberFixedImagesField.setValue(properties.getNrFixedImages());
        numberWritableImagesField.setValue(properties.getNrWritableImages());
        slideShowCheck.setSelected(properties.hasSlideShow());
        if (properties.getRgbColors() != null) {
            rgbColorsCheck.setSelected(true);
            fixedColorPaletteCheck.setSelected(false);
            rgbColorsField.setText("");
            fixedColorPaletteField.setEditable(false);
        } else {
            rgbColorsCheck.setSelected(false);
            fixedColorPaletteCheck.setSelected(true);
            rgbColorsField.setEditable(false);
            fixedColorPaletteField.setText("");
        }
        pngSupportedCheck.setSelected(properties.isPngSupported());
        textDisplayRowsField.setText(Integer.toString(properties.getTextDisplayRows()));
        textDisplayColumnsField.setText(Integer.toString(properties.getTextDisplayColumns()));
    }

    /**
     * @return the softwareVersionField
     */
    public final JTextField getSoftwareVersionField() {
        return softwareVersionField;
    }

    /**
     * @return the displayTypeField
     */
    public final JComboBox<ETdpDisplayType> getDisplayTypeField() {
        return displayTypeField;
    }

    /**
     * @return the protocolField
     */
    public final NumericTextField getProtocolField() {
        return protocolField;
    }

    /**
     * @return the supplierAndProductField
     */
    public final JTextField getSupplierAndProductField() {
        return supplierAndProductField;
    }

    /**
     * @return the serialField
     */
    public final JTextField getSerialField() {
        return serialField;
    }

    /**
     * @return the externalLightingCheck
     */
    public final JCheckBox getExternalLightingCheck() {
        return externalLightingCheck;
    }

    /**
     * @return the displayHeightField
     */
    public final NumericTextField getDisplayHeightField() {
        return displayHeightField;
    }

    /**
     * @return the displayWidthField
     */
    public final NumericTextField getDisplayWidthField() {
        return displayWidthField;
    }

    /**
     * @return the numberFixedImagesField
     */
    public final NumericTextField getNumberFixedImagesField() {
        return numberFixedImagesField;
    }

    /**
     * @return the numberWritableImagesField
     */
    public final NumericTextField getNumberWritableImagesField() {
        return numberWritableImagesField;
    }

    /**
     * @return the slideShowCheck
     */
    public final JCheckBox getSlideShowCheck() {
        return slideShowCheck;
    }

    /**
     * @return the rgbColorsCheck
     */
    public final JCheckBox getRgbColorsCheck() {
        return rgbColorsCheck;
    }

    /**
     * @return the rgbColorsField
     */
    public final JTextField getRgbColorsField() {
        return rgbColorsField;
    }

    /**
     * @return the fixedColorPaletteCheck
     */
    public final JCheckBox getFixedColorPaletteCheck() {
        return fixedColorPaletteCheck;
    }

    /**
     * @return the fixedColorPaletteField
     */
    public final JTextField getFixedColorPaletteField() {
        return fixedColorPaletteField;
    }

    /**
     * @return the pngSupportedCheck
     */
    public final JCheckBox getPngSupportedCheck() {
        return pngSupportedCheck;
    }

    /**
     * @return the textDisplayRowsField
     */
    public NumericTextField getTextDisplayRowsField() {
        return textDisplayRowsField;
    }

    /**
     * @return the textDisplayColumnsField
     */
    public NumericTextField getTextDisplayColumnsField() {
        return textDisplayColumnsField;
    }
}
