/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
   list of conditions and the following disclaimer in the documentation and/or
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 */
package org.disperanto.displaysimulator.ui.panels;

import java.awt.event.ActionListener;
import java.awt.event.ItemListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import net.miginfocom.swing.MigLayout;

import org.disperanto.data.TdpStatus;
import org.disperanto.displaysimulator.listeners.IStatusListener;
import org.disperanto.displaysimulator.ui.components.NumericTextField;
import org.disperanto.enums.ETdpLightingState;

/**
 * Panel for showing current status of a display.
 */
public class StatusPanel extends JPanel implements IStatusListener {

    private static final boolean HAS_LIGHT_SENSORS = false;
    private static final boolean HAS_TEMPERATURE_SENSOR = true;
    private static final boolean HAS_HEATING_SENSOR = true;
    private static final boolean HAS_COOLING_SENSOR = true;

    // components
    private JLabel shownImagesField;
    private NumericTextField brightnessField;
    private NumericTextField intensityExternalLightingField;
    private JCheckBox lightSensorsCheck;
    private NumericTextField measuredLightIntensity1Field;
    private NumericTextField measuredLightIntensity2Field;
    private JTextField gpsPositionField;
    private JCheckBox temperatureCheck;
    private NumericTextField temperatureField;
    private JCheckBox heatingCheck;
    private JComboBox<ETdpLightingState> heatingField;
    private JCheckBox coolingCheck;
    private JComboBox<ETdpLightingState> coolingField;

    private JButton saveButton;

    /**
     * Constructor
     */
    public StatusPanel() {
        initUI();
    }

    /**
     * Add listener for panel actions
     *
     * @param listener Action listener
     */
    public final void addActionListener(ActionListener listener) {
        saveButton.addActionListener(listener);
    }

    /**
     * Add listener for item actions
     *
     * @param listener Action listener
     */
    public final void addItemListener(ItemListener listener) {
        lightSensorsCheck.addItemListener(listener);
        temperatureCheck.addItemListener(listener);
        heatingCheck.addItemListener(listener);
        coolingCheck.addItemListener(listener);
    }

    private void initUI() {
        setLayout(new MigLayout());

        setBorder(BorderFactory.createTitledBorder("Status"));

        shownImagesField = new JLabel();
        brightnessField = new NumericTextField(1, 100);
        intensityExternalLightingField = new NumericTextField(1, 100);
        lightSensorsCheck = new JCheckBox();
        measuredLightIntensity1Field = new NumericTextField(0, 100);
        measuredLightIntensity2Field = new NumericTextField(0, 100);
        gpsPositionField = new JTextField(10);
        temperatureCheck = new JCheckBox();
        temperatureField = new NumericTextField(0, 100);
        heatingCheck = new JCheckBox();
        heatingField = new JComboBox<ETdpLightingState>(ETdpLightingState.values());
        heatingField.removeItem(ETdpLightingState.AUTOMATIC); // do not show automatic option
        coolingCheck = new JCheckBox();
        coolingField = new JComboBox<ETdpLightingState>(ETdpLightingState.values());
        coolingField.removeItem(ETdpLightingState.AUTOMATIC); // do not show automatic option
        saveButton = new JButton("Apply");

        add(new JLabel("Shown image(s):"));
        add(shownImagesField, "wrap");

        add(new JLabel("Brightness (%):"));
        add(brightnessField, "wrap, grow");

        add(new JLabel("Intensity external lighting (%):"));
        add(intensityExternalLightingField, "wrap, grow");

        add(new JLabel("Light sensors:"));
        add(lightSensorsCheck, "wrap, grow");

        add(new JLabel("Measured light intensity 1 (%):"));
        add(measuredLightIntensity1Field, "wrap, grow");

        add(new JLabel("Measured light intensity 2 (%):"));
        add(measuredLightIntensity2Field, "wrap, grow");

        add(new JLabel("GPS position:"));
        add(gpsPositionField, "wrap, grow");

        add(new JLabel("Temperature:"));
        add(temperatureCheck, "split 2");
        add(temperatureField, "wrap, grow");

        add(new JLabel("Heating:"));
        add(heatingCheck, "split 2");
        add(heatingField, "wrap, grow");

        add(new JLabel("Cooling:"));
        add(coolingCheck, "split 2");
        add(coolingField, "wrap, grow");

        add(new JLabel(""));
        add(saveButton, "wrap");
    }

    @Override
    public final void statusChanged(int displayAddress, TdpStatus status) {
        if (status.getShownImages().size() > 0) {
            shownImagesField.setText(status.getShownImages().toString());
        } else {
            shownImagesField.setText("No image shown");
        }
        brightnessField.setValue(status.getBrightnessInPercentage());
        if (status.getIntensityExternalLighting() == null) {
            intensityExternalLightingField.setValue(1);
            intensityExternalLightingField.setEnabled(false);
        } else {
            intensityExternalLightingField.setValue(status.getIntensityExternalLighting());
            intensityExternalLightingField.setEnabled(true);
        }
        if (status.getMeasuredLightIntensity().size() > 0) {
            lightSensorsCheck.setSelected(true);
        } else {
            lightSensorsCheck.setSelected(false);
        }
        gpsPositionField.setText(status.getGpsPosition());
        temperatureCheck.setSelected(HAS_TEMPERATURE_SENSOR);
        if (HAS_TEMPERATURE_SENSOR) {
            temperatureCheck.setSelected(true);
            temperatureField.setValue(status.getTemperature());
        } else {
            temperatureCheck.setSelected(true);
        }
        if (HAS_HEATING_SENSOR) {
            heatingCheck.setSelected(true);
            heatingField.setSelectedItem(status.getHeating());
        } else {
            heatingCheck.setSelected(false);
        }
        if (HAS_COOLING_SENSOR) {
            coolingCheck.setSelected(true);
            coolingField.setSelectedItem(status.getHeating());
        } else {
            coolingCheck.setSelected(false);
        }
    }

    /**
     * @return the hasLightSensors
     */
    public final boolean isHasLightSensors() {
        return HAS_LIGHT_SENSORS;
    }

    /**
     * @return the hasTemperatureSensor
     */
    public final boolean isHasTemperatureSensor() {
        return HAS_TEMPERATURE_SENSOR;
    }

    /**
     * @return the hasHeatingSensor
     */
    public final boolean isHasHeatingSensor() {
        return HAS_HEATING_SENSOR;
    }

    /**
     * @return the hasCoolingSensor
     */
    public final boolean isHasCoolingSensor() {
        return HAS_COOLING_SENSOR;
    }

    /**
     * @return the shownImagesField
     */
    public final JLabel getShownImagesField() {
        return shownImagesField;
    }

    /**
     * @return the brightnessField
     */
    public final NumericTextField getBrightnessField() {
        return brightnessField;
    }

    /**
     * @return the intensityExternalLightingField
     */
    public final NumericTextField getIntensityExternalLightingField() {
        return intensityExternalLightingField;
    }

    /**
     * @return the lightSensorsCheck
     */
    public final JCheckBox getLightSensorsCheck() {
        return lightSensorsCheck;
    }

    /**
     * @return the measuredLightIntensity1Field
     */
    public final NumericTextField getMeasuredLightIntensity1Field() {
        return measuredLightIntensity1Field;
    }

    /**
     * @return the measuredLightIntensity2Field
     */
    public final NumericTextField getMeasuredLightIntensity2Field() {
        return measuredLightIntensity2Field;
    }

    /**
     * @return the gpsPositionField
     */
    public final JTextField getGpsPositionField() {
        return gpsPositionField;
    }

    /**
     * @return the temperatureCheck
     */
    public final JCheckBox getTemperatureCheck() {
        return temperatureCheck;
    }

    /**
     * @return the temperatureField
     */
    public final NumericTextField getTemperatureField() {
        return temperatureField;
    }

    /**
     * @return the heatingCheck
     */
    public final JCheckBox getHeatingCheck() {
        return heatingCheck;
    }

    /**
     * @return the heatingField
     */
    public final JComboBox<ETdpLightingState> getHeatingField() {
        return heatingField;
    }

    /**
     * @return the coolingCheck
     */
    public final JCheckBox getCoolingCheck() {
        return coolingCheck;
    }

    /**
     * @return the coolingField
     */
    public final JComboBox<ETdpLightingState> getCoolingField() {
        return coolingField;
    }
}
