package org.disperanto.displaysimulator.data;

import static org.disperanto.enums.ETdpCommunicationError.ERROR_UNKNOWN_COMMAND;
import static org.disperanto.enums.ETdpNotification.COMMUNICATION_ERROR;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpNotification;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpNotification;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.NotificationResponse;
import org.disperanto.messages.TdpMessage;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class MessageHandlerTest {

    private List<Display> displays;
    private Display firstDisplay;
    private Display unknownDisplay;
    private ILogger logger;
    
    @Before
    public void before() {
        DisplayCreator displayCreator = new DisplayCreator();
        firstDisplay = displayCreator.createDisplay(1);
        unknownDisplay = displayCreator.createDisplay(2);

        displays = new ArrayList<>();
        displays.add(firstDisplay);
        // unknownDisplay is not included in the list of displays

        logger = new TestLogger();
    }

    /**
     * Test behavior when receiving an unsupported command type
     * Tests fix for #MM-3932
     */
    @Test
    public void handleUnknownCommand() throws TdpProtocolException {
        TdpMessage response = handleCommand(createCommand(firstDisplay, ETdpCommand.UNKNOWN));
        // Expect ERROR_UNKNOWN_COMMAND notification
        Assert.assertEquals(ETdpCommand.NOTIFICATION, response.getCommand());
        List<TdpNotification> notifications = NotificationResponse.parse(response);
        Assert.assertEquals(1, notifications.size());
        assertNotificationError(notifications.get(0), COMMUNICATION_ERROR, ERROR_UNKNOWN_COMMAND);
    }

    /** Checks behavior when receiving a command for an unknown address */
    @Test
    public void handleUnknownAddress() {
        TdpMessage command = createCommand(unknownDisplay, ETdpCommand.KEEP_ALIVE);
        List<TdpMessage> responses = MessageHandler.handleCommands(displays, Arrays.asList(command), logger);
        // Unknown addresses are ignored -- no responses
        Assert.assertEquals(Collections.emptyList(), responses);
    }

    private void assertNotificationError(TdpNotification ntf, ETdpNotification ntfType, ETdpCommunicationError error) {
        Assert.assertEquals(ntfType, ntf.getTag());

        Assert.assertTrue(ntf instanceof TdpCommunicationError);
        TdpCommunicationError errorNtf = (TdpCommunicationError)ntf;
        Assert.assertEquals(error, errorNtf.getCommunicationError());
    }

    private TdpMessage createCommand(Display target, ETdpCommand commandType) {
        return new TdpMessage(ETdpMessageType.COMMAND, commandType, Arrays.asList(target.getAddress()),
                ByteBuffer.allocate(0));
    }

    private TdpMessage handleCommand(TdpMessage... commands) {
        List<TdpMessage> responses = MessageHandler.handleCommands(displays, Arrays.asList(commands), logger);
        Assert.assertEquals(1, responses.size());
        return responses.get(0);
    }
}
