/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.utils;

/**
 * Utils for hot keys.
 */
public abstract class HotkeyUtils {

    /**
     * Strips the ampersand (&amp;) from a string. For example "Test &button" returns "Test button".
     * 
     * @param text Text to handle
     * @return Stripped text
     */
    public static String stripMnemonic(String text) {
        if (text == null) {
            return null;
        }
        final StringBuffer sb = new StringBuffer();
        for (String s : text.split("&")) {
            // If the ampersand is followed by a hash (#), then it is ignored. This allows for HTML codes such as
            // &amp;#235;
            if (s.length() > 0) {
                if (s.charAt(0) == '#') {
                    sb.append("&").append(s);
                } else {
                    sb.append(s);
                }
            }
        }
        return sb.toString();
    }

    /**
     * Checks if the specified text has a mnemonic (i.e. a character preceded by an ampersand &amp;).
     * 
     * @param text Text to handle
     * @return true if text contains mnemonic, otherwise false
     */
    public static boolean hasMnemonic(String text) {
        if (text == null) {
            return false;
        }
        return !stripMnemonic(text).equals(text);
    }

    /**
     * Gets the character to be used as a mnemonic (hotkey). For example "Test &button" returns 'B'.
     * 
     * @param text Text to handle
     * @return Hot key
     */
    public static char getMnemonic(String text) {
        if (text == null) {
            return (char)0;
        }
        boolean skipFirst = true;
        for (String s : text.split("&")) {
            if (skipFirst) {
                skipFirst = false;
            } else {
                // If the ampersand is followed by a hash (#), then it is ignored. This allows for HTML codes such as
                // &amp;#235;
                if ((s.length() > 0) && (s.charAt(0) != '#')) {
                    return Character.toUpperCase(s.charAt(0));
                }
            }
        }
        return (char)0;
    }
}
