/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;

import org.disperanto.enums.EIpConnectionType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * TCP sender to send messages and receive the responses.
 * 
 * The TcpSender is a TCP client that connects to a TcpReceiver.
 */
public class TcpSender implements IStartableConnection {

    private static final Logger LOGGER = LoggerFactory.getLogger(TcpSender.class);

    private final String ipAddress;
    private final int port;
    private final int readTimeoutMs;
    private Socket socket;

    /**
     * Creates a new TcpSender connecting to the TcpReceiver on specified address and port.
     *  
     * @param ipAddress IP address of the TcpReceiver.
     * @param port port number of the TCP receiver.
     */
    public TcpSender(String ipAddress, int port, int readTimeoutMs) {
        this.ipAddress = ipAddress;
        this.port = port;
        this.readTimeoutMs = readTimeoutMs;
    }

    @Override
    public void send(ByteBuffer data) throws IOException {
        if (socket == null) {
            // if the initial start failed, then try to start it again. 
            start();
        }
        
        if (socket != null && !socket.isClosed()) {
            OutputStream out = socket.getOutputStream();
            out.write(data.array());
            out.flush();
        }
    }

    @Override
    public InputStream getInputStream() throws IOException {
        if (socket != null) {
            return socket.getInputStream();
        }
        return null;
    }
    
    @Override
    public EIpConnectionType getConnectionType() {
        return EIpConnectionType.TCP;
    }

    @Override
    public String getIpAddress() {
        return ipAddress;
    }

    @Override
    public int getPort() {
        return port;
    }

    @Override
    public void start() throws UnknownHostException, IOException {
        socket = new Socket(ipAddress, port);
        socket.setSoTimeout(readTimeoutMs);
    }

    @Override
    public void stop() {
        if (socket != null) {
            try {
                socket.close();
                socket = null;
            } catch (IOException e) {
                LOGGER.error("Unable to close socket to " + ipAddress, e);
            }
        }
    }
}
