/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;

import org.disperanto.enums.EIpConnectionType;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.TdpUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Sets up a UDP connection to send commands and receive responses.
 */
public class UdpSender implements IStartableConnection {

    private static final Logger LOGGER = LoggerFactory.getLogger(UdpSender.class);

    // Address of receiver of UDP message
    private String ipAddress;
    // resolved IP address
    private InetAddress inetAddress;
    // Port used by receiver of UDP message
    private int port;
    private int timeoutMs;

    private DatagramSocket socket;

    /**
     * Constructor
     *
     * @param ipAddress IP address of the receiver
     * @param port Port used sending UDP packets
     * @param timeoutMs Timeout for UDP communication in milliseconds
     * @throws UnknownHostException when ipAddress cannot be resolved.
     */
    public UdpSender(String ipAddress, int port, int timeoutMs) throws UnknownHostException {
        setIpAddress(ipAddress);
        this.port = port;
        this.timeoutMs = timeoutMs;
    }

    /**
     * Starts the UDP connection.
     */
    @Override
    public final void start() {
        try {
            socket = new DatagramSocket();
            socket.setSoTimeout(timeoutMs);
        } catch (SocketException ex) {
            LOGGER.trace("SocketException: " + ex.getMessage());
        }
    }

    /**
     * Stops the UDP connection.
     */
    @Override
    public final void stop() {
        if (socket != null) {
            socket.close();
            socket = null;
        }
    }

    @Override
    public final void send(ByteBuffer data) throws IOException {
        try {
            data.rewind();
            final DatagramPacket packet =
                    new DatagramPacket(data.array(), data.capacity(), inetAddress, port);
            socket.send(packet);

            LOGGER.trace("Send packet: " + packet.getLength() + " bytes to " + packet.getAddress() + ":" +
                    packet.getPort() + ": " + ByteBufferUtils.toByteString(data));
        } catch (IllegalArgumentException ex) {
            LOGGER.trace("IllegalArgumentException: " + ex.getMessage());
            throw new IOException("Unresolved address: " + ipAddress);
        }
    }

    /**
     * Receives a UDP packet from the same address a packet is sent to. <br>
     * Packets from other IP addresses are skipped (no handling at all).
     *
     * @return UDP packet
     * @throws IOException
     */
    private DatagramPacket receivePacket() throws IOException {
        boolean incorrectIpAddress = true;

        final ByteBuffer packetData = ByteBuffer.allocate(TdpUtils.MAX_BUFFER);

        final DatagramPacket packet = new DatagramPacket(packetData.array(), packetData.capacity());

        while (incorrectIpAddress) {
            socket.receive(packet);

            // check if packet is received from expected IP address
            if (!packet.getAddress().equals(inetAddress)) {
                // packet received from unexpected IP address
                LOGGER.warn("Packet received from unexpected IP-address " + packet.getAddress().toString() +
                        ". Expected is " + inetAddress);

                incorrectIpAddress = true;
            } else {
                incorrectIpAddress = false;
            }
        }

        return packet;
    }

    @Override
    public EIpConnectionType getConnectionType() {
        return EIpConnectionType.UDP;
    }

    @Override
    public String getIpAddress() {
        return ipAddress;
    }

    @Override
    public int getPort() {
        return port;
    }

    public void setPort(int port) {
        this.port = port;
    }

    public int getTimeoutMs() {
        return timeoutMs;
    }

    public void setTimeoutMs(int timeoutMs) {
        this.timeoutMs = timeoutMs;
    }

    public final void setIpAddress(String ipAddress) throws UnknownHostException {
        this.ipAddress = ipAddress;
        inetAddress = InetAddress.getByName(ipAddress);
    }

    @Override
    public InputStream getInputStream() throws IOException {
        ByteBuffer buffer = receive();
        if (buffer != null && buffer.hasArray()) {
            buffer.rewind();
            return new ByteArrayInputStream(buffer.array());
        }
        return null;
    }

    private ByteBuffer receive() throws IOException {
        final DatagramPacket packet = receivePacket();

        final ByteBuffer messageData = ByteBuffer.allocate(packet.getLength());
        messageData.put(packet.getData(), 0, packet.getLength());

        LOGGER.trace("Received packet: " + packet.getLength() + " bytes: " + ByteBufferUtils.toByteString(messageData));
        return messageData;
    }
}
