/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;

import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.exception.TdpProtocolException;


/**
 * Formatter for the SetBrightnessTable command.
 */
public final class SetBrightnessCommand {
    private static final int NR_ENTRIES = 11;

    /**
     * Constructor
     */
    private SetBrightnessCommand() {
        // Empty constructor
    }

    /**
     * Creates the data buffer for the SetBrightnessTable command.<br>
     *
     * The array list with the brightness percentages must have 11 entries (0% - 100% with increments of 10%).
     *
     * @param brightnessTable Data of this command message
     * @return Data buffer for the brightness percentages
     */
    public static ByteBuffer createData(TdpBrightnessTable brightnessTable) {
        final ByteBuffer tableData = ByteBuffer.allocate(NR_ENTRIES);

        if (brightnessTable.getBrightnessTable().size() != NR_ENTRIES) {
            throw new IllegalArgumentException(
                    "The brightness table must contain exactly 11 entries (0% - 100% with increments of 10%)");
        }

        for (Integer entry : brightnessTable.getBrightnessTable()) {
            tableData.put(setBrightnessPercentage(entry.intValue()));
        }

        return tableData;
    }

    /**
     * Parses the received data to a TdpBrightnessTable object <br>
     *
     * The byte buffer with the brightness percentages contains 11 bytes (0% - 100% with increments of 10%).
     *
     * @param message Generic message that is received from a display
     * @return Brightness table of the Disperanto
     */
    public static TdpBrightnessTable parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final TdpBrightnessTable brightnessTable = new TdpBrightnessTable();
        final ArrayList<Integer> brightnessList = new ArrayList<Integer>();

        formatter.checkCommand(message, ETdpCommand.SET_BRIGHTNESS);

        // Parse data
        final ByteBuffer data = formatter.getMessageData(message);

        if (data.capacity() != 11) {
            throw new TdpProtocolException(
                    "The brightness table must contain exactly 11 entries (0% - 100% with increments of 10%)");
        }

        for (int i = 0; i < 11; i++) {
            brightnessList.add(parseBrightnessPercentage(data));
        }
        brightnessTable.setBrightnessTable(brightnessList);

        formatter.checkRemainingData(data);

        return brightnessTable;
    }

    /**
     * Parses the data for a brightness percentage.
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Brightness percentage
     * @throws TdpProtocolException
     */
    private static int parseBrightnessPercentage(final ByteBuffer data) throws TdpProtocolException {
        final int percentage = data.get();

        if (percentage < 0 || percentage > 100) {
            throw new TdpProtocolException(
                    "Incorrect value. The entries in the brightness table indicate a percentage. "
                            + "Value must be between [0 - 100]");
        }

        return percentage;
    }

    private static ByteBuffer setBrightnessPercentage(int percentage) {
        final TdpFormatter formatter = new TdpFormatter();

        if (percentage < 0 || percentage > 100) {
            throw new IllegalArgumentException(
                    "Incorrect value. The values in the brightness table indicate a percentage. "
                            + "Value must be between [0 - 100]");
        }

        return formatter.setOneByteData(percentage, "brightness percentage", ETdpCommand.SET_BRIGHTNESS.getName());
    }
}
