/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpExternalLighting;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpLightingState;
import org.disperanto.exception.TdpProtocolException;


/**
 * Formatter for the SetExternalLighting command.
 */
public final class SetExternalLightingCommand {
    /**
     * Constructor
     */
    private SetExternalLightingCommand() {
        // Empty constructor
    }

    /**
     * Creates the data buffer for the SetExternalLighting command.
     *
     * @param timeoutSettings Data of this command message
     * @return Data buffer for the timeout settings
     */
    public static ByteBuffer createData(TdpExternalLighting timeoutSettings) {
        final ByteBuffer settingsData = ByteBuffer.allocate(1);

        settingsData.put((byte)timeoutSettings.getState().getValue());

        return settingsData;
    }

    /**
     * Parses the received data to a TdpExternalLighting object
     *
     * @param message Generic message that is received from a display
     * @return Timeout settings of the Disperanto
     */
    public static TdpExternalLighting parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final TdpExternalLighting lightingSettings = new TdpExternalLighting();

        formatter.checkCommand(message, ETdpCommand.SET_EXTERNAL_LIGHTING);

        // Parse data
        final ByteBuffer data = formatter.getMessageData(message);

        lightingSettings.setState(parseState(data));

        formatter.checkRemainingData(data);

        return lightingSettings;
    }

    /**
     * Parses the data for the communication timeout mode.<br>
     * 
     * Only the first data of the buffer is read.
     * 
     * @param data Byte buffer containing the data to be parsed
     * @return Communication timeout mode
     * @throws TdpProtocolException
     */
    private static ETdpLightingState parseState(ByteBuffer data) throws TdpProtocolException {
        final ETdpLightingState state = ETdpLightingState.convertValue(data.get());

        if (state != null) {
            return state;
        } else {
            throw new TdpProtocolException("Unknown value for state in SetExternalLighting command");
        }
    }
}
