/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpTimeoutSettings;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpTimeoutMode;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.VlqUtils;


/**
 * Formatter for the SetCommunicationTimeout command.
 */
public final class SetTimeoutCommand {
    private static final String MESSAGENAME = ETdpCommand.SET_COMMUNICATION_TIMEOUT.getName();

    /**
     * Constructor
     */
    private SetTimeoutCommand() {
        // Empty constructor
    }

    /**
     * Creates the data buffer for the SetCommunicationTimeout command.
     *
     * @param timeoutSettings Data of this command message
     * @return Data buffer for the timeout settings
     */
    public static ByteBuffer createData(TdpTimeoutSettings timeoutSettings) {
        final ByteBuffer settingsData = ByteBuffer.allocate(1 + (VlqUtils.MAX_VLQ_BYTES * 2));

        settingsData.put(setTimeoutModeData(timeoutSettings));

        if (timeoutSettings.getTimeoutMode() != ETdpTimeoutMode.NO_TIMEOUT) {
            settingsData.put(setTimeoutValueData(timeoutSettings));
        }

        if (timeoutSettings.getTimeoutMode() == ETdpTimeoutMode.SHOW_IMAGE) {
            settingsData.put(VlqUtils.getVlqBytes(timeoutSettings.getImageSlot()));
        }

        return ByteBufferUtils.truncateAtCurrentPosition(settingsData);
    }

    /**
     * Parses the received data to a TdpTimeoutSettings object
     *
     * @param message Generic message that is received from a display
     * @return Timeout settings of the Disperanto
     */
    public static TdpTimeoutSettings parse(TdpMessage message) throws TdpProtocolException {
        final TdpTimeoutSettings timeoutSettings = new TdpTimeoutSettings();
        final TdpFormatter formatter = new TdpFormatter();

        formatter.checkCommand(message, ETdpCommand.SET_COMMUNICATION_TIMEOUT);

        // Parse data
        final ByteBuffer data = formatter.getMessageData(message);

        timeoutSettings.setTimeoutMode(parseTimeoutModeData(data));

        if (timeoutSettings.getTimeoutMode() != ETdpTimeoutMode.NO_TIMEOUT) {
            timeoutSettings.setTimeoutInSec(VlqUtils.parseVlqData(data));
        }

        if (timeoutSettings.getTimeoutMode() == ETdpTimeoutMode.SHOW_IMAGE) {
            timeoutSettings.setImageSlot(VlqUtils.parseVlqData(data));
        }

        formatter.checkRemainingData(data);

        return timeoutSettings;
    }

    /**
     * Parses the data for the communication timeout mode.<br>
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Communication timeout mode
     * @throws TdpProtocolException
     */
    private static ETdpTimeoutMode parseTimeoutModeData(ByteBuffer data) throws TdpProtocolException {
        final ETdpTimeoutMode timeoutMode = ETdpTimeoutMode.convertValue(data.get());

        if (timeoutMode != null) {
            return timeoutMode;
        } else {
            throw new TdpProtocolException("Unknown value for timeout mode in SetCommunicationTimeout command");
        }
    }

    /**
     * Sets the data for the communication timeout mode.<br>
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Communication timeout mode
     * @throws TdpProtocolException
     */
    private static ByteBuffer setTimeoutModeData(TdpTimeoutSettings settings) {
        final TdpFormatter formatter = new TdpFormatter();

        if (settings.getTimeoutMode() == null) {
            throw new IllegalArgumentException("Timeout mode is mandatory in SetCommunicationTimeout command");
        }

        return formatter.setOneByteData(settings.getTimeoutMode().getValue(), "timeout mode", MESSAGENAME);
    }

    /**
     * Sets the data for the communication timeout value (in seconds).<br>
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Communication timeout mode
     * @throws TdpProtocolException
     */
    private static ByteBuffer setTimeoutValueData(TdpTimeoutSettings settings) {
        if (settings.getTimeoutMode() != ETdpTimeoutMode.NO_TIMEOUT) {
            if (settings.getTimeoutInSec() <= 0) {
                throw new IllegalArgumentException("Timeout value is mandatory when timeout mode not is NO_TIMEOUT " +
                        "in SetCommunicationTimeout command");
            }
        }

        return VlqUtils.getVlqBytes(settings.getTimeoutInSec());
    }
}
