/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpSlideShowMode;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.VlqUtils;


/**
 * Formatter for the StartSlideShow command.
 */
public final class StartSlideShowCommand {
    /**
     * Constructor
     */
    private StartSlideShowCommand() {
        // Empty constructor
    }

    /**
     * Creates the data buffer for the StartSlideShow command.
     * 
     * @param settings Data of this command message
     * @return Data buffer for the timeout settings
     */
    public static ByteBuffer createData(TdpStartSlideShow settings) {
        final int bufferSize = (settings.getImageSlotsSetting().size() * 2 * VlqUtils.MAX_VLQ_BYTES) + 1;
        final ByteBuffer settingsData = ByteBuffer.allocate(bufferSize);

        settingsData.put((byte)settings.getMode().getValue());

        for (Entry<Integer, Integer> slide : settings.getImageSlotsSetting().entrySet()) {
            settingsData.put(VlqUtils.getVlqBytes(slide.getKey()));
            settingsData.put(VlqUtils.getVlqBytes(slide.getValue()));
        }

        return ByteBufferUtils.truncateAtCurrentPosition(settingsData);
    }

    /**
     * Parses the received data to a TdpStartSlideShow object.
     *
     * @param message Generic Disperanto message that is received
     * @return Slide show mode
     */
    public static TdpStartSlideShow parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final TdpStartSlideShow slideShowSettings = new TdpStartSlideShow();

        formatter.checkCommand(message, ETdpCommand.START_SLIDE_SHOW);

        // Parse data
        final ByteBuffer data = formatter.getMessageData(message);

        slideShowSettings.setMode(parseSlideShowMode(data));

        slideShowSettings.setImageSlotsSetting(parseSlideShowSettings(data));

        return slideShowSettings;
    }

    /**
     * Parses the data for the slide show mode.
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Slide show mode
     * @throws TdpProtocolException
     */
    private static ETdpSlideShowMode parseSlideShowMode(ByteBuffer data) throws TdpProtocolException {
        final ETdpSlideShowMode slideShowMode = ETdpSlideShowMode.convertValue(data.get());

        if (slideShowMode != null) {
            return slideShowMode;
        } else {
            throw new TdpProtocolException("Unknown value for slide show mode in StartSlideShow command");
        }
    }

    /**
     * Parses the data for the slide show settings (list of image slots and time to show image in tenth of a second).
     * 
     * @param data Byte buffer containing the data to be parsed
     * @return Slide show settings
     * @throws TdpProtocolException
     */
    private static LinkedHashMap<Integer, Integer> parseSlideShowSettings(ByteBuffer data)
            throws TdpProtocolException {
        // use LinkedHashMap to preserve order.
        final LinkedHashMap<Integer, Integer> settingsMap = new LinkedHashMap<Integer, Integer>();

        while (data.hasRemaining()) {
            final int imageslot = VlqUtils.parseVlqData(data);
            final int showTime = VlqUtils.parseVlqData(data);
            settingsMap.put(Integer.valueOf(imageslot), Integer.valueOf(showTime));
        }

        if (settingsMap.isEmpty()) {
            // no images set for slide show
            throw new TdpProtocolException("No images set for slide show");
        }
        return settingsMap;
    }
}
