/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpDiagnostics;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpExternalLighting;
import org.disperanto.data.TdpManipulateMemorySlot;
import org.disperanto.data.TdpNotification;
import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpShowImage;
import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.data.TdpStatus;
import org.disperanto.data.TdpTimeoutSettings;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpNotification;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Base class for creating messages.
 */
public final class TdpCommandFactory {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpCommandFactory.class);

    /**
     * Constructor
     */
    private TdpCommandFactory() {
        // Empty constructor
    }

    /**
     * Creates a DisplayProperties command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing DisplayProperties command
     */
    public static TdpMessage createDisplayPropertiesCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, addresses,
                null);

        LOGGER.debug("Display properties command created: " + message.toString());

        return message;
    }

    /**
     * Creates a DisplayProperties response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param displayProperties Properties of this display
     * @return Disperanto message containing DisplayProperties response
     */
    public static TdpMessage createDisplayPropertiesResponse(int address, int messageNumber,
            TdpDisplayProperties displayProperties) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.DISPLAY_PROPERTIES,
                Arrays.asList(address), DisplayPropertiesResponse.createData(displayProperties));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("Display properties response created: " + message.toString());

        return message;
    }

    /**
     * Creates a Status command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing Status command
     */
    public static TdpMessage createStatusCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.STATUS, addresses, null);

        LOGGER.debug("Status command created: " + message.toString());

        return message;
    }

    /**
     * Creates a Status response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param status Status variables of this display
     * @return Disperanto message containing Status response
     */
    public static TdpMessage createStatusResponse(int address, int messageNumber, TdpStatus status) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.STATUS, Arrays.asList(address),
                StatusResponse.createData(status));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("Status response created: " + message.toString());

        return message;
    }

    /**
     * Creates a Reboot command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing Reboot command
     */
    public static TdpMessage createRebootCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.REBOOT, addresses, null);

        LOGGER.debug("Reboot command created: " + message.toString());

        return message;
    }

    /**
     * Creates a Reboot response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing Reboot response
     */
    public static TdpMessage createRebootResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.REBOOT, Arrays.asList(address),
                null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("Reboot response created: " + message.toString());

        return message;
    }

    /**
     * Creates a KeepAlive command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing KeepAlive command
     */
    public static TdpMessage createKeepAliveCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.KEEP_ALIVE, addresses, null);

        LOGGER.debug("KeepAlive command created: " + message.toString());

        return message;
    }

    /**
     * Creates a KeepAlive response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing KeepAlive response
     */
    public static TdpMessage createKeepAliveResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.KEEP_ALIVE,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("KeepAlive response created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetCommunicationTimeout command.
     *
     * @param addresses Display addresses the command is sent to
     * @param settings Timeout settings variables for the displays
     * @return Disperanto message containing SetCommunicationTimeout command
     */
    public static TdpMessage createSetCommunicationTimeoutCommand(List<Integer> addresses,
            TdpTimeoutSettings settings) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SET_COMMUNICATION_TIMEOUT,
                addresses, SetTimeoutCommand.createData(settings));

        LOGGER.debug("SetCommunicationTimout command created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetCommunicationTimeout response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing SetCommunicationTimeout response
     */
    public static TdpMessage createSetCommunicationTimeoutResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SET_COMMUNICATION_TIMEOUT,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("SetCommunicationTimout response created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetBrightnessTable command.
     *
     * @param addresses Display addresses the command is sent to
     * @param table Brightness table for the displays
     * @return Disperanto message containing SetBrightnessTable command
     */
    public static TdpMessage createSetBrightnessTableCommand(List<Integer> addresses, TdpBrightnessTable table) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SET_BRIGHTNESS, addresses,
                SetBrightnessCommand.createData(table));

        LOGGER.debug("SetBrightnessTable command created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetBrightnessTable response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing SetBrightnessTable response
     */
    public static TdpMessage createSetBrightnessTableResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SET_BRIGHTNESS,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("SetBrightnessTable response created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetExternalLighting command.
     *
     * @param addresses Display addresses the command is sent to
     * @param settings External lighting settings for the displays
     * @return Disperanto message containing SetExternalLighting command
     */
    public static TdpMessage createSetExternalLightingCommand(List<Integer> addresses, TdpExternalLighting settings) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SET_EXTERNAL_LIGHTING,
                addresses, SetExternalLightingCommand.createData(settings));

        LOGGER.debug("SetExternalLighting command created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetExternalLighting response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing SetExternalLighting response
     */
    public static TdpMessage createSetExternalLightingResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SET_EXTERNAL_LIGHTING,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("SetExternalLighting response created: " + message.toString());

        return message;
    }

    /**
     * Creates a Diagnostics command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing Diagnostics command
     */
    public static TdpMessage createDiagnosticsCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DIAGNOSTICS, addresses, null);

        LOGGER.debug("Diagnostics command created: " + message.toString());

        return message;
    }

    /**
     * Creates a Diagnostics response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param diagnostics diagnostic data of the display
     * @return Disperanto message containing Diagnostics response
     */
    public static TdpMessage createDiagnosticsResponse(int address, int messageNumber, TdpDiagnostics diagnostics) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.DIAGNOSTICS,
                Arrays.asList(address), DiagnosticsResponse.createData(diagnostics));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("Diagnostics response created: " + message.toString());

        return message;
    }

    /**
     * Creates a ManipulateMemorySlot command.
     *
     * @param addresses Display addresses the command is sent to
     * @param data Data for manipulate memory slot for the displays
     * @return Disperanto message containing ManipulateMemorySlot command
     */
    public static TdpMessage createManipulateMemorySlotCommand(List<Integer> addresses, TdpManipulateMemorySlot data) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.MANIPULATE_MEMORY_SLOT,
                addresses, ManipulateMemorySlotCommand.createData(data));

        LOGGER.debug("ManipulateMemorySlot command created: " + message.toString());

        return message;
    }

    /**
     * Creates a ManipulateMemorySlot response.
     * 
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param crc CRC of the working memory
     * @return Disperanto message containing ManipulateMemorySlot response
     */
    public static TdpMessage createManipulateMemorySlotResponse(int address, int messageNumber, short crc) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.MANIPULATE_MEMORY_SLOT,
                Arrays.asList(address), ManipulateMemorySlotResponse.createData(crc));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("ManipulateMemorySlot response created: " + message.toString());

        return message;
    }

    /**
     * Creates a CalculateCrcImages command.
     *
     * @param addresses Display addresses the command is sent to
     * @param imageSlots list of image slots for which the crc is requested
     * @return Disperanto message containing CalculateCrcImages command
     */
    public static TdpMessage createCalculateCrcImagesCommand(List<Integer> addresses, List<Integer> imageSlots) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.CALCULATE_CRC_IMAGES, addresses,
                CalculateCrcImagesCommand.createData(imageSlots));

        LOGGER.debug("CalculateCrcImages command created: " + message.toString());

        return message;
    }

    /**
     * Creates a CalculateCrcImages response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param crcs List of crc's of the requested image slots
     * @return Disperanto message containing CalculateCrcImages response
     */
    public static TdpMessage createCalculateCrcImagesResponse(int address, int messageNumber, List<Short> crcs) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.CALCULATE_CRC_IMAGES,
                Arrays.asList(address), CalculateCrcImagesResponse.createData(crcs));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("CalculateCrcImages response created: " + message.toString());

        return message;
    }

    /**
     * Creates a ShowNoImage command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing ShowNoImage command
     */
    public static TdpMessage createShowNoImageCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SHOW_NO_IMAGE, addresses, null);

        LOGGER.debug("ShowNoImage command created: " + message.toString());

        return message;
    }

    /**
     * Creates a ShowNoImage response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing ShowNoImage response
     */
    public static TdpMessage createShowNoImageResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SHOW_NO_IMAGE,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("ShowNoImage response created: " + message.toString());

        return message;
    }

    /**
     * Creates a ShowImage command.
     *
     * @param addresses Display addresses the command is sent to
     * @param data ShowImage Data containing image slot to display
     * @return Disperanto message containing ShowImage command
     */
    public static TdpMessage createShowImageCommand(List<Integer> addresses, TdpShowImage data) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SHOW_IMAGE, addresses,
                ShowImageCommand.createData(data));

        LOGGER.debug("ShowImage command created: " + message.toString());

        return message;
    }

    /**
     * Creates a ShowImage response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing ShowImage response
     */
    public static TdpMessage createShowImageResponse(int address, int messageNumber, short crc) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SHOW_IMAGE,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        message.setData(ShowImageResponse.createData(crc));

        LOGGER.debug("ShowImage response created: " + message.toString());

        return message;
    }

    /**
     * Creates a StartSlideShow command.
     *
     * @param addresses Display addresses the command is sent to
     * @param data Data containing information for the slide show to display
     * @return Disperanto message containing StartSlideShow command
     */
    public static TdpMessage createStartSlideShowCommand(List<Integer> addresses, TdpStartSlideShow data) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.START_SLIDE_SHOW, addresses,
                StartSlideShowCommand.createData(data));

        LOGGER.debug("StartSlideShow command created: " + message.toString());

        return message;
    }

    /**
     * Creates a StartSlideShow response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @param crcs List of crc's of the images requested by the slideshow
     * @return Disperanto message containing StartSlideShow response
     */
    public static TdpMessage createStartSlideShowResponse(int address, int messageNumber, List<Short> crcs) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.START_SLIDE_SHOW,
                Arrays.asList(address),  StartSlideShowResponse.createData(crcs));
        message.setMessageNumber(messageNumber);

        LOGGER.debug("StartSlideShow response created: " + message.toString());

        return message;
    }

    /**
     * Creates a ServiceMode command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing ServiceMode command
     */
    public static TdpMessage createServiceModeCommand(List<Integer> addresses) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SUPPLIER_SERVICE_MODE,
                addresses, null);

        LOGGER.debug("ServiceMode command created: " + message.toString());

        return message;
    }

    /**
     * Creates a ServiceMode response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing ServiceMode response
     */
    public static TdpMessage createServiceModeResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SUPPLIER_SERVICE_MODE,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("ServiceMode command created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetText command.
     *
     * @param addresses Display addresses the command is sent to
     * @return Disperanto message containing SetText command
     */
    public static TdpMessage createSetTextCommand(List<Integer> addresses, TdpSetText data) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.SET_TEXT, addresses,
                SetTextCommand.createData(data));

        LOGGER.debug("SetText command created: " + message.toString());

        return message;
    }

    /**
     * Creates a SetText response.
     *
     * @param address Display address that sends the response
     * @param messageNumber Message number (should be copied from command)
     * @return Disperanto message containing SetText response
     */
    public static TdpMessage createSetTextResponse(int address, int messageNumber) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.SET_TEXT,
                Arrays.asList(address), null);
        message.setMessageNumber(messageNumber);

        LOGGER.debug("SeText command created: " + message.toString());

        return message;
    }

    /**
     * Creates a ClearNotifications command.
     *
     * @param addresses Display addresses for which the command is sent
     * @return Disperanto message containing ServiceMode command
     */
    public static TdpMessage createClearNotificationsCommand(List<Integer> addresses,
            Collection<ETdpNotification> notifications) {

        ByteBuffer buffer = ByteBuffer.allocate(notifications.size());
        for (ETdpNotification notification : notifications) {
            buffer.put((byte)notification.getValue());
        }

        final TdpMessage message =
                new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.NOTIFICATION, addresses, buffer);

        LOGGER.debug("Notification command created: " + message.toString());

        return message;
    }

    /**
     * Creates a Notification response.
     *
     * @param address Display address that sends the notification
     * @param notifications List of notification from this display
     * @return Disperanto message containing ServiceMode command
     */
    public static TdpMessage createNotification(int address, List<TdpNotification> notifications) {
        final TdpMessage message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.NOTIFICATION,
                Arrays.asList(address), NotificationResponse.createData(notifications));

        LOGGER.debug("Notification response created: " + message.toString());

        return message;
    }

    /** Creates a notification to report a {@link TdpCommunicationError} */
    public static TdpMessage createErrorNotification(int address, ETdpCommunicationError error) {
        return createNotification(address, Arrays.asList(new TdpCommunicationError(error)));
    }

}
