/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.VlqUtils;


/**
 * Base class for all message formatters.
 *
 * Contains the basic functions for parsing and setting tlv data in a message.
 * Can create empty messages.
 */
public class TdpFormatter {
    /**
     * Creates a formatter for Disperanto messages
     */
    public TdpFormatter() {
        // Empty constructor
    }

    /**
     * Parses one byte of the data for the given tag of the given message.<br>
     *
     * @param data Byte buffer containing the data to be parsed
     * @param dataDescription Description of the data that is read
     * @param messageDescription Description of the message
     * @return Unsigned value that is parsed
     */
    protected final int parseOneByteData(ByteBuffer data, String dataDescription, String messageDescription)
            throws TdpProtocolException {
        // expected data length 1
        if (data.capacity() != 1) {
            throw new TdpProtocolException("Wrong data length received for " + dataDescription + " tag from " +
                    messageDescription + " response");
        }

        data.rewind();
        return (data.get() & 0xFF);
    }

    /**
     * Parses one byte of the data for the given tag of the given message.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param dataDescription Description of the data that is read
     * @param messageDescription Description of the message
     * @return Signed value that is parsed
     */
    protected final int parseOneByteDataSigned(ByteBuffer data, String dataDescription, String messageDescription)
            throws TdpProtocolException {
        // expected data length 1
        if (data.capacity() != 1) {
            throw new TdpProtocolException("Wrong data length received for " + dataDescription + " tag from " +
                    messageDescription + " response");
        }

        data.rewind();
        return (data.get());
    }

    /**
     * Parses the data (1 or 2 bytes) for the given tag of the given message. <br>
     * Data contains VLQ-coded quantity.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param dataDescription Description of the data that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    protected final int parseMaxTwoBytesData(ByteBuffer data, String dataDescription, String messageDescription)
            throws TdpProtocolException {
        // expected data length is max 2 bytes
        if (data.capacity() > 2) {
            throw new TdpProtocolException("Wrong data length received for " + dataDescription + " tag from " +
                    messageDescription + " response");
        }

        return VlqUtils.parseVlqData(data);
    }

    /**
     * Parses the byte buffer data for the given tag of the given message.<br>
     * <br>
     * Only length is checked. Data can be returned without parsing if length check is ok.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param dataLength Expected data length of the data that is parsed
     * @param dataDescription Description of the data that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    protected final ByteBuffer parseByteBufferData(ByteBuffer data, int dataLength, String dataDescription,
            String messageDescription) throws TdpProtocolException {
        // check data length
        if (data.capacity() != dataLength) {
            throw new TdpProtocolException("Wrong data length received for " + dataDescription + " tag from " +
                    messageDescription + " response");
        }

        data.rewind();
        return data;
    }

    /**
     * Parses the data of a string with the given max length for the given tag of the given message.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param length Length of the string that is parsed
     * @param stringDescription Description of the string that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    protected final String parseStringData(ByteBuffer data, int length, String stringDescription,
            String messageDescription) throws TdpProtocolException {
        // check max length of string
        if (data.capacity() > length) {
            throw new TdpProtocolException("Wrong data length received for " + stringDescription + " tag from " +
                    messageDescription + " response");
        }

        data.rewind();
        return ByteBufferUtils.byteBufferToString(data);
    }

    /**
     * parses the data for a percentage value in a byte buffer.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param tagDescription Description of the tag of the tlv data that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    protected final Integer parsePercentageData(ByteBuffer data, String tagDescription, String messageDescription)
            throws TdpProtocolException {
        final int value = parseOneByteData(data, tagDescription, messageDescription);

        if ((value < 0) || (value > 100)) {
            throw new TdpProtocolException("Unexpected value for " + tagDescription + " data of " + messageDescription +
                    " message (" + value + "). Expected is percentage (0..100).");
        }

        return Integer.valueOf(value);
    }

    /**
     * Sets the data for the brightness in a byte buffer.
     *
     * @param value value of the brightness of the display
     * @param tagDescription Description of the tag of the tlv data that is read
     * @param messageDescription Description of the message
     */
    protected final ByteBuffer setPercentageData(int value, String tagDescription, String messageDescription) {
        if ((value < 0) || (value > 100)) {
            throw new IllegalArgumentException("Unexpected value for " + tagDescription + " data of " +
                    messageDescription + " message (" + value + "). Expected is percentage (0..100).");
        }

        return setOneByteData(value, tagDescription, messageDescription);
    }

    /**
     * Sets one byte of data in a byte buffer.
     *
     * @param value Value that is put in the byte buffer
     * @param valueDescription Description of the value that is read
     * @param messageDescription Description of the message
     * @return Byte buffer with given value
     */
    protected final ByteBuffer setOneByteData(int value, String valueDescription, String messageDescription) {
        final ByteBuffer data = ByteBuffer.allocate(1);

        if (value > 0xFF) {
            throw new IllegalArgumentException("Value for " + valueDescription + " of " + messageDescription +
                    " does not fit in one byte");
        }
        data.put((byte)value);

        data.rewind();
        return data;
    }

    /**
     * Sets one or two bytes in a byte buffer depending on value.
     *
     * @param value Value that is put in the byte buffer
     * @param valueDescription Description of the value that is read
     * @param messageDescription Description of the message
     * @return Byte buffer with given value
     */
    protected final ByteBuffer setMaxTwoBytesData(int value, String valueDescription, String messageDescription) {
        final ByteBuffer data = ByteBuffer.allocate(VlqUtils.countVlqBytes(value));

        data.put(VlqUtils.getVlqBytes(value));

        data.rewind();
        return data;
    }

    /**
     * Sets the data for the given string in a byte buffer.
     *
     * @param string String that is put in a byte buffer
     * @param maxLength Maximum length of given string
     * @param valueDescription Description of the value that is read
     * @param messageDescription Description of the message
     * @return Byte buffer with given value
     */
    protected final ByteBuffer setStringData(String string, int maxLength, String valueDescription,
            String messageDescription) {
        final ByteBuffer data = ByteBuffer.allocate(string.length());

        if (string.length() > maxLength) {
            throw new IllegalArgumentException("String length for " + valueDescription + " of " + messageDescription +
                    " is too long (" + string.length() + "). Max length = " + maxLength);
        }
        data.put(ByteBufferUtils.stringToByteBuffer(string));

        data.rewind();
        return data;
    }

    /**
     * Sets the byte buffer data.<br>
     * <br>
     * Only length is checked. Data can be returned without parsing if length check is ok.
     *
     * @param data Data of a byte buffer
     * @param maxLength Maximum length of given string
     * @param dataDescription Description of the data that is read
     * @param messageDescription Description of the message
     * @return Byte buffer with given value
     */
    protected final ByteBuffer setByteBufferData(ByteBuffer data, int maxLength, String dataDescription,
            String messageDescription) {
        // expected data length is 3 bytes
        if (data.capacity() != maxLength) {
            throw new IllegalArgumentException("Data length for " + dataDescription + " of " + messageDescription +
                    " is incorrect (" + data.capacity() + "). Expected length = " + maxLength);
        }

        data.rewind();
        return data;
    }

    /**
     * Gets the message data and checks if the data is present.
     *
     * @param message Disperanto message
     * @return Byte buffer containing the data of the Disperanto message
     */
    protected final ByteBuffer getMessageData(TdpMessage message) throws TdpProtocolException {
        final ByteBuffer data = message.getData();
        if ((data == null)) {
            throw new TdpProtocolException("No data in message");
        }

        data.rewind();
        return data;
    }

    /**
     * checks if there is no data left in the data buffer of the Disperanto message
     *
     * @param data Byte buffer containing the message data of a Disperanto message, with position in
     *            buffer at last read byte.
     */
    protected final void checkRemainingData(final ByteBuffer data) throws TdpProtocolException {
        if (data.hasRemaining()) {
            throw new TdpProtocolException("Too much data received for message");
        }
    }

    /**
     * Checks if the retrieved message is a response.
     *
     * @param response Message that is retrieved
     * @param commandId expected command id in message
     */
    protected final void checkResponse(TdpMessage response, ETdpCommand commandId) throws TdpProtocolException {
        if (response.getMessageType() != ETdpMessageType.RESPONSE) {
            throw new TdpProtocolException("Wrong message type in response (" + response.getMessageType().getName() +
                    ").");
        }

        if (response.getCommand() != commandId) {
            throw new TdpProtocolException("Wrong command id in response (expected = " + commandId.getName() +
                    "; observed = " + response.getCommand().getName() + ")");
        }

        if (response.getAddresses().size() != 1) {
            throw new TdpProtocolException("Wrong number of addresses in response (" + response.getAddresses().size() +
                    "). Should always be 1");
        }
    }

    /**
     * Checks if the retrieved message is a command.
     *
     * @param command Message that is retrieved
     * @param commandId expected command id in message
     */
    protected final void checkCommand(TdpMessage command, ETdpCommand commandId) throws TdpProtocolException {
        if (command.getMessageType() != ETdpMessageType.COMMAND) {
            throw new TdpProtocolException("Wrong message type in command (" + command.getMessageType().getName() +
                    ").");
        }

        if (command.getCommand() != commandId) {

            throw new TdpProtocolException("Wrong command id in command (expected = " + commandId.getName() +
                    "; observed = " + command.getCommand().getName() + ")");
        }
    }
}
