/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.protocol;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.communication.IConnection;
import org.disperanto.data.TdpNotifications;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.NotificationResponse;
import org.disperanto.messages.TdpMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Sends Disperanto commands and waits for the response (timeout used).
 */
public class TdpProtocolExchanger extends TdpProtocol {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpProtocolExchanger.class);

    private final List<ITdpNotificationListener> notificationListeners = new ArrayList<ITdpNotificationListener>();
    private final IConnection connection;

    /**
     * Constructor
     *
     * @param connection UDP connection
     */
    public TdpProtocolExchanger(IConnection connection) {
        this.connection = connection;
    }

    /**
     * Exchanges a list of Traffic Display messages.<br>
     * <br>
     * A list of Traffic Display commands is sent to one or more displays.<br>
     * Traffic Display responses and notifications are received from each display that has received a command.<br>
     * This function keeps reading until all expected data is received or a timeout is elapsed.
     *
     * @param messages List of Traffic Display commands for the sign group
     * @return List of Traffic Display responses and notifications sent by the sign group
     */
    public final List<TdpMessage> exchangeMessages(List<TdpMessage> messages) throws IOException, TdpProtocolException {
        sendMessages(connection, messages);

        LOGGER.debug("Start receiving");
        List<TdpMessage> responses = receiveMessages(connection);

        if (responses.isEmpty()) {
            LOGGER.debug("No list with TDP responses received");
            throw new TdpProtocolException("No response received");
        }

        LOGGER.debug("Check response received");
        handleNotifications(responses);
        return responses;
    }

    /**
     * Notifications are moved to another list and handled by calling the notification listeners.
     *
     * @param responses Warning! this method alters this list
     * @throws TdpProtocolException
     * @throws IOException
     */
    private void handleNotifications(List<TdpMessage> responses) throws TdpProtocolException, IOException {
        final List<TdpMessage> notificationMessages = new ArrayList<TdpMessage>();
        final TdpNotifications notifications = new TdpNotifications();

        if (responses == null) {
            // no messages received --> no notifications to handle
            return;
        }

        for (TdpMessage message : responses) {
            // only one display address expected in a response
            notifications.setDisplayAddress(message.getAddresses().get(0));
            if (message.getCommand() == ETdpCommand.NOTIFICATION) {
                notifications.addNotifications(NotificationResponse.parse(message));
                notificationMessages.add(message);
            }
        }

        responses.removeAll(notificationMessages);

        notify(notifications);
    }

    /**
     * Adds a listener for received commands.
     *
     * @param listener Listener for received commands
     */
    public final void addNotificationListener(ITdpNotificationListener listener) {
        notificationListeners.add(listener);
    }

    /**
     * Removes a listener for received commands.
     *
     * @param listener Listener for received commands
     */
    public final void removeNotificationListener(ITdpNotificationListener listener) {
        notificationListeners.remove(listener);
    }

    /**
     * Notifies a notification to the listeners. Notifications are removed from the list.
     *
     * @throws TdpProtocolException
     * @throws IOException
     */
    private void notify(TdpNotifications notifications) throws TdpProtocolException, IOException {
        for (ITdpNotificationListener listener : notificationListeners) {
            listener.notificationReceived(notifications);
        }
    }
}
