/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.utils;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;

import org.disperanto.enums.ETdpDataLength;
import org.disperanto.exception.TdpProtocolException;


/**
 * Reads the compact TLV (Tag Length Value) data from a given data buffer, element by element.
 */
public class CompactTlvReader {
    private final ByteBuffer tlvData;
    private int currentId;
    private ByteBuffer currentData;

    /**
     * Constructor
     * 
     * @param tlvData The TLV (Tag Length Value) data
     */
    public CompactTlvReader(ByteBuffer tlvData) {
        this.tlvData = tlvData;
        tlvData.rewind();
    }

    /**
     * Reads the next tag and the data that belongs to this tag from the buffer.
     *
     * @return false if no more data is available in the buffer
     */
    public final boolean next() throws TdpProtocolException {
        if (!tlvData.hasRemaining()) {
            return false;
        }

        // read the tag
        final byte tag = tlvData.get();

        // read data
        currentData = readTlvData(tag, tlvData);

        // get id from the tag (bits 0-5)
        currentId = (byte)(tag & 0x3F);

        return true;
    }

    /**
     * @return the currentId
     */
    public final int getId() {
        return currentId;
    }

    /**
     * @return the currentData
     */
    public final ByteBuffer getData() {
        currentData.rewind();
        return currentData;
    }

    /**
     * Gets the data length from the tag.<br>
     *
     * Bits 6 and 7 contain the data length:
     * <ul>
     * <li>bit 7 = 0, bit 6 = 0: no data
     * <li>bit 7 = 0, bit 6 = 1: 1 data byte
     * <li>bit 7 = 1, bit 6 = 0: 2 data bytes
     * <li>bit 7 = 1, bit 6 = 1: data length defined in VLQ-bytes
     * </ul>
     * <br>
     *
     * @param tag Byte containing the data length
     * @param data Data that is being read
     * @return Data length
     */
    protected final int getTlvDataLength(byte tag, ByteBuffer data) throws TdpProtocolException {
        // Get data length from tag
        int dataLength = (tag & 0xC0) >> 6;
        if (dataLength == ETdpDataLength.VLQ_USED.getValue()) {
            // read VLQ-bytes to determine data length
            dataLength = VlqUtils.parseVlqData(data);
        }

        return dataLength;
    }

    /**
     * Reads the compact TLV data.
     *
     * @param tag First byte of TLV data
     * @param messageData Received data buffer
     * @return Compact TLV data
     * @throws TdpProtocolException
     */
    private ByteBuffer readTlvData(byte tag, ByteBuffer messageData) throws TdpProtocolException {
        final int length = getTlvDataLength(tag, messageData);
        final ByteBuffer data = TdpUtils.createBuffer(length);

        // read data
        try {
            for (int i = 0; i < length; i++) {
                data.put(messageData.get());
            }
        } catch (BufferUnderflowException e) {
            throw new TdpProtocolException("Not all expected data received in message");
        }

        return data;
    }
}
