/*
Copyright (c) 2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.utils;

import java.io.ByteArrayOutputStream;
import java.nio.ByteBuffer;

import org.disperanto.enums.ETdpDataLength;


/**
 * Writes the compact TLV (Tag Length Value) data for a given data buffer, element by element.
 */
public class CompactTlvWriter {
    private final ByteArrayOutputStream tlvOut; 

    /**
     * Constructor
     *
     * @param initialCapacity The initial buffer capacity
     */
    public CompactTlvWriter(int initialCapacity) {
        tlvOut = new ByteArrayOutputStream(initialCapacity);
    }

    /**
     * Writes the next tag with the data of the tag to the buffer.
     *
     * @param tag Identification of the data
     * @param data data of the tag
     */
    public final void add(byte tag, ByteBuffer data) {
        // set the tag and length
        setTagAndLength(tag, data);

        // write data
        if (data != null) {
            data.rewind();
            for (int n = 0; n < data.limit(); n++) {
                tlvOut.write(data.get(n));
            }
        }
    }

    /**
     * Sets the tag and the length indication.<br>
     *
     * In case the data length is more than two bytes, VLQ byte(s) are added for the data length.
     *
     * @param tag Byte containing the identification and data length indicator
     * @param data Data that is being read
     */
    protected final void setTagAndLength(byte tag, ByteBuffer data) {
        if (data == null) {
            tlvOut.write((byte)(tag | (ETdpDataLength.NO_DATA.getValue() << 6)));
        } else if (data.capacity() <= 2) {
            tlvOut.write((byte)(tag | (data.capacity() << 6)));
        } else {
            tlvOut.write((byte)(tag | (ETdpDataLength.VLQ_USED.getValue() << 6)));

            // add VLQ byte
            ByteBuffer buf = VlqUtils.getVlqBytes(data.capacity());
            buf.rewind();
            for (int n = 0; n < buf.limit(); n++) {
                tlvOut.write(buf.get(n));
            }
        }
    }
    
    /**
     * @return The compact TLV (Tag Length Value) data for a given data buffer
     */
    public ByteBuffer getData() {
        return ByteBuffer.wrap(tlvOut.toByteArray());
    }
}
