/*
Copyright (c) 2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.utils;

import java.nio.ByteBuffer;

import org.disperanto.exception.TdpProtocolException;


/**
 * Generic utilities for handling the Variable Length Quantity (VLQ).<br>
 *
 * The value is stored in a variable number of bytes. The highest bit of every byte indicates whether there are more
 * bytes. The other 7 bits represent the value. The order of the bytes is big endian. The most significant bits are
 * first in the stream.<br>
 *
 * Values up to 127 fit in one byte. Values from 128 up to 16383 are stored in two bytes. Values from 16384 up to
 * 2097151 are stored in 3 bytes.
 */
public abstract class VlqUtils {

    // The maximum value for VLQ values in this protocol is 2 pow 31-1 (2147483647), 5 bytes are needed
    public static final int MAX_VALUE = 2147483647;
    public static final int MAX_VLQ_BYTES = 5;

    /**
     * Calculates the number of VLQ bytes that are needed to store the quantity.<br>
     *
     * Each VLQ byte can store 7 bits of the value. The highest bit is used to indicate whether there are more VLQ
     * bytes.
     *
     * @param value Value that is handled
     * @return Number of VLQ bytes needed to store the value
     */
    public static int countVlqBytes(final int value) {
        int val = value;

        if (value < 0) {
            throw new IllegalArgumentException("No negative value expected for VLQ quantity");
        }

        if (value == 0) {
            // 0 is stored in 1 VLQ-byte
            return 1;
        }

        // Get the number of bits needed for the value
        int bitCount = 0;
        while (val > 0) {
            bitCount++;
            val = val >> 1;
        }

        // get number of VLQ-bytes
        return (int)Math.ceil(bitCount / 7.0);
    }

    /**
     * Parses the quantity from the VLQ bytes of the received data.<br>
     * Bit 7 indicates if there are more VLQ bytes. Bit 7 = 1, there are more VLQ bytes.
     *
     * @param data Received byte buffer marked just before the position of the VLQ bytes
     * @return quantity Value in the VLQ bytes
     */
    public static int parseVlqData(ByteBuffer data) throws TdpProtocolException {
        int quantity = 0;

        int vlqByte = data.get();

        // check if there are more vlq-bytes (bit 7 = 1)
        while (((vlqByte & 0x80) >> 7) == 1) {
            quantity = (quantity << 7) + (vlqByte & 0x7F);

            if (data.hasRemaining()) {
                vlqByte = data.get();
            } else {
                throw new TdpProtocolException("No more data bytes while more VLQ-byte(s) are expected");
            }
        }
        // add last byte
        quantity = (quantity << 7) + (vlqByte & 0x7F);
        return quantity;
    }

    /**
     * Checks whether or not this is the last VLQ byte.
     *
     * @param vlqByte The byte to check.
     * @return <code>true</code> when this is the last byte, <code>false</code> otherwise.
     */
    public static boolean hasMoreVLQBytes(int vlqByte) {
        return ((vlqByte & 0x80) >> 7) == 1;
    }

    /**
     * Puts the quantity in VLQ bytes.<br>
     *
     * Bit 7 indicates if there are more VLQ bytes. If bit 7 = 1, then there are more VLQ bytes.
     *
     * @param quantity Value that is stored in one or more VLQ bytes
     * @return VLQ-bytes
     */
    public static ByteBuffer getVlqBytes(int quantity) {

        final int nrVlqBytes = VlqUtils.countVlqBytes(quantity);
        int length = quantity;

        final ByteBuffer vlq = ByteBuffer.allocate(nrVlqBytes);

        for (int i = nrVlqBytes; i > 0; i--) {
            // get the last seven bytes
            final int sevenBits = length & 0x7F;

            length = length >> 7;
            // set these bits in the byte buffer and add a flag in the highest bit to indicate if there are more
            // vlq-bytes
            if (i == nrVlqBytes) {
                vlq.put(i - 1, (byte)sevenBits);
            } else {
                // last byte of data length: bit 7 is set to 1
                vlq.put(i - 1, (byte)(sevenBits | 0x80));
            }
        }

        return vlq;
    }
}
