/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.List;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpStatus;
import org.disperanto.refapp.prisserver.TdpMobiMaestroService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for a display controller.<br>
 *
 * Make sure a DisplayController is running.<br>
 *
 * This application sends requests to the display controller and logs the responses.
 */
public final class RunTestDisplayController {
    private static final Logger LOGGER = LoggerFactory.getLogger(RunTestDisplayController.class);

    private static TdpMobiMaestroService server;
    private static int displayAddress = 6;

    private RunTestDisplayController() {
        // Empty constructor
    }

    /**
     * Starts server.
     * 
     * @param ipAddress IP address of display controller
     * @param port Port used to retrieve message
     */
    public static void init(String ipAddress, int port) throws UnknownHostException {
        server = new TdpMobiMaestroService(ipAddress, port, 1000);
        server.start();
    }

    /**
     * Stops server.
     */
    public static void clean() {
        // Stops the display controller
        server.stop();
    }

    /**
     * Tests the request of display properties.
     */
    public static void testDisplayProperties() {
        try {
            final List<TdpDisplayProperties> displayPropertiesList =
                    server.requestDisplayProperties(Arrays.asList(displayAddress));

            if (displayPropertiesList == null) {
                throw new Exception("No response received for Display Properties command");
            }

            for (TdpDisplayProperties properties : displayPropertiesList) {
                LOGGER.debug("Display properties: " + properties.toString());
            }
        } catch (Exception ex) { // SUPPRESS CHECKSTYLE IllegalCatchCheck
            ex.printStackTrace();
        }
    }

    /**
     * Tests the request of status info.
     */
    public static void testStatus() {
        try {
            final List<TdpStatus> statusList = server.requestStatus(Arrays.asList(displayAddress));

            if (statusList == null) {
                throw new Exception("No response received for Status command");
            }

            for (TdpStatus status : statusList) {
                LOGGER.debug("Status: " + status.toString());
            }
        } catch (Exception ex) { // SUPPRESS CHECKSTYLE IllegalCatchCheck
            ex.printStackTrace();
        }
    }

    /**
     * Starts the test.
     */
    public static void main(String[] args) throws Exception {
        String ipAddress;
        int port;

        // handle arguments
        if (args.length != 3) {
            System.out.println("Usage: 4 arguments are mandatory:\nIP-address of display controller, "
                    + "port used by display controller, display address");

            return;
        } else {
            ipAddress = args[0];
            port = Integer.valueOf(args[1]).intValue();
            displayAddress = Integer.valueOf(args[2]).intValue();
        }

        // start server
        init(ipAddress, port);

        // send requests
        testDisplayProperties();
        testStatus();

        // stop server
        clean();
    }
}
