/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpNotification;
import org.disperanto.data.TdpStatus;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpDisplayType;
import org.disperanto.enums.ETdpNotification;
import org.disperanto.refapp.displaycontroller.Display;
import org.disperanto.refapp.displaycontroller.TdpDisplayControllerService;
import org.disperanto.refapp.prisserver.TdpMobiMaestroService;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * JUnit tests for the class TdpProtocol.
 */
@Ignore(value = "Disabled because unit tests aren't allowed to open sockets")
public class SystemTest {
    private static final Logger LOGGER = LoggerFactory.getLogger(SystemTest.class);

    private TdpMobiMaestroService server;
    private TdpDisplayControllerService client;
    
    /**
     * Default setup for each test.
     */
    @Before
    public final void setUp() throws IOException {
        server = new TdpMobiMaestroService("127.0.0.1", 7890, 1000);
        server.start();
        client = new TdpDisplayControllerService(null, 7890, 1000);
        client.start();
    }

    /**
     * Default setup for each test.
     */
    @After
    public final void cleanUp() {
        // Stops the display controller
        client.stop();
        server.stop();
    }

    /**
     * Tests the request of display properties.
     */
    @Test
    public final void testDisplayProperties() {
        final int address = 21;
        try {
            // add display with display properties
            final TdpDisplayProperties displayProperties = new TdpDisplayProperties();
            displayProperties.setProtocolVersion(255);
            displayProperties.setDisplayType(ETdpDisplayType.MATRIX);
            displayProperties.setSupplierAndProductType("1234567890123456789012345678901234567890");
            displayProperties.setSerialNumber("12345678901234567890");
            displayProperties.setSoftwareVersion("12345678901234567890");
            displayProperties.setExternalLighting(true);
            displayProperties.setDisplayHeightInPixels(0x01);
            displayProperties.setDisplayWidthInPixels(0x1000);
            displayProperties.setNrFixedImages(128);
            displayProperties.setNrWritableImages(16383);
            displayProperties.setSlideShow(true);
            displayProperties.setNrSlideShowImages(3);
            final ByteBuffer rgbColors = ByteBuffer.allocate(3);
            rgbColors.put((byte)0xFF);
            rgbColors.put((byte)0xFF);
            rgbColors.put((byte)0xFF);
            displayProperties.setRgbColors(rgbColors);
            displayProperties.setPngSupported(true);

            final Display display = new Display(address);
            display.setDisplayProperties(displayProperties);
            display.addNotifications(new TdpNotification(ETdpNotification.COLD_RESTART));
            client.addDisplay(display);

            final List<TdpDisplayProperties> displayPropertiesList =
                    server.requestDisplayProperties(Arrays.asList(address));

            Assert.assertEquals(1, displayPropertiesList.size());

            for (TdpDisplayProperties properties : displayPropertiesList) {
                LOGGER.info(properties.toString());
                Assert.assertEquals(display.getDisplayProperties().getProtocolVersion(),
                        properties.getProtocolVersion());
                Assert.assertEquals(display.getDisplayProperties().getDisplayType(), properties.getDisplayType());
                Assert.assertEquals(display.getDisplayProperties().getSupplierAndProductType(),
                        properties.getSupplierAndProductType());
                Assert.assertEquals(display.getDisplayProperties().getSerialNumber(), properties.getSerialNumber());
                Assert.assertEquals(display.getDisplayProperties().getSoftwareVersion(),
                        properties.getSoftwareVersion());
                Assert.assertEquals(display.getDisplayProperties().hasExternalLighting(),
                        properties.hasExternalLighting());
                Assert.assertEquals(display.getDisplayProperties().getDisplayHeightInPixels(),
                        properties.getDisplayHeightInPixels());
                Assert.assertEquals(display.getDisplayProperties().getDisplayWidthInPixels(),
                        properties.getDisplayWidthInPixels());
                Assert.assertEquals(display.getDisplayProperties().getNrFixedImages(), properties.getNrFixedImages());
                Assert.assertEquals(display.getDisplayProperties().getNrWritableImages(),
                        properties.getNrWritableImages());
                Assert.assertEquals(display.getDisplayProperties().hasSlideShow(), properties.hasSlideShow());
                Assert.assertEquals(display.getDisplayProperties().getNrSlideShowImages(),
                        properties.getNrSlideShowImages());
                Assert.assertEquals(display.getDisplayProperties().getRgbColors().rewind(), properties.getRgbColors());
                Assert.assertEquals(display.getDisplayProperties().getFixedColorPalette(),
                        properties.getFixedColorPalette());
                Assert.assertEquals(display.getDisplayProperties().isPngSupported(), properties.isPngSupported());
            }
        } catch (Exception ex) { // SUPPRESS CHECKSTYLE IllegalCatchCheck
            ex.printStackTrace();
            Assert.fail();
        }
    }

    /**
     * Tests the request of the status.
     */
    @Test
    public final void testStatus() {
        final int address = 21;
        try {
            // add display with display properties
            final TdpStatus statusOriginal = new TdpStatus();
            final HashMap<Integer, Short> shownImages = new HashMap<Integer, Short>();
            shownImages.put(1, (short)0x1234);
            shownImages.put(6, (short)0x9876);
            statusOriginal.setShownImages(shownImages);
            statusOriginal.setBrightnessInPercentage(12);
            statusOriginal.setIntensityExternalLighting(26);
            final ArrayList<Integer> measuredLightIntensity = new ArrayList<Integer>();
            measuredLightIntensity.add(Integer.valueOf(0));
            measuredLightIntensity.add(Integer.valueOf(50));
            measuredLightIntensity.add(Integer.valueOf(100));
            statusOriginal.setMeasuredLightIntensity(measuredLightIntensity);
            statusOriginal.setGpsPosition("N52.987655 E5.123456");
            statusOriginal.setTemperature(-5);
            statusOriginal.setHeating(1);
            statusOriginal.setCooling(0);

            final Display display = new Display(address);
            display.setStatus(statusOriginal);
            display.addNotifications(new TdpCommunicationError(ETdpCommunicationError.ERROR_UNKNOWN_COMMAND));
            display.addNotifications(new TdpCommunicationError(ETdpCommunicationError.ERROR_ILLEGAL_DATA));
            display.addNotifications(new TdpNotification(ETdpNotification.EXTERNAL_LIGHTING_DEFECT));
            client.addDisplay(display);

            final List<TdpStatus> statusList = server.requestStatus(Arrays.asList(address));

            Assert.assertEquals(1, statusList.size());

            for (TdpStatus status : statusList) {
                LOGGER.info(status.toString());
                Assert.assertEquals(display.getStatus().getShownImages(), status.getShownImages());
                Assert.assertEquals(display.getStatus().getBrightnessInPercentage(),
                        status.getBrightnessInPercentage());
                Assert.assertEquals(display.getStatus().getIntensityExternalLighting(),
                        status.getIntensityExternalLighting());
                Assert.assertEquals(display.getStatus().getMeasuredLightIntensity(),
                        status.getMeasuredLightIntensity());
                Assert.assertEquals(display.getStatus().getGpsPosition(), status.getGpsPosition());
                Assert.assertEquals(display.getStatus().getTemperature(), status.getTemperature());
                Assert.assertEquals(display.getStatus().getHeating(), status.getHeating());
                Assert.assertEquals(display.getStatus().getCooling(), status.getCooling());
            }
        } catch (Exception ex) { // SUPPRESS CHECKSTYLE IllegalCatchCheck
            ex.printStackTrace();
            Assert.fail();
        }
    }
}
