/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class CalculateCrcImagesCommand.
 */
public class TdpCommandCalculateCrcImagesTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse() {
        try {
            final ByteBuffer data = TdpUtils.createBuffer(7);
            data.put(VlqUtils.getVlqBytes(127));
            data.put(VlqUtils.getVlqBytes(1));
            data.put(VlqUtils.getVlqBytes(128));
            data.put(VlqUtils.getVlqBytes(16384));

            // correct message
            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.CALCULATE_CRC_IMAGES);
            message.addAddress(1);
            message.setData(data);
            final List<Integer> imageSlotList = CalculateCrcImagesCommand.parse(message);
            Assert.assertEquals(127, imageSlotList.get(0).intValue());
            Assert.assertEquals(1, imageSlotList.get(1).intValue());
            Assert.assertEquals(128, imageSlotList.get(2).intValue());
            Assert.assertEquals(16384, imageSlotList.get(3).intValue());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function createData()
     */
    @Test
    public final void testCreateData() {
        try {
            // correct message
            final List<Integer> list = Arrays.asList(1, 127, 128, 16383, 16384);
            final ByteBuffer data = CalculateCrcImagesCommand.createData(list);
            data.rewind();
            Assert.assertEquals(1, VlqUtils.parseVlqData(data));
            Assert.assertEquals(127, VlqUtils.parseVlqData(data));
            Assert.assertEquals(128, VlqUtils.parseVlqData(data));
            Assert.assertEquals(16383, VlqUtils.parseVlqData(data));
            Assert.assertEquals(16384, VlqUtils.parseVlqData(data));
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong response id
     */
    @Test
    public final void testParseExceptionWrongCommandId() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in command"));

        // incorrect message (wrong response id)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        CalculateCrcImagesCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseExceptionWrongMessageType() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in command"));

        // incorrect message (response)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.CALCULATE_CRC_IMAGES);
        message.addAddress(1);
        CalculateCrcImagesCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseExceptionNoDataNull() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.CALCULATE_CRC_IMAGES);
        message.addAddress(1);
        message.setData(null);
        CalculateCrcImagesCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseExceptionNoDataEmpty() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.CALCULATE_CRC_IMAGES);
        message.addAddress(1);
        message.setData(null);
        CalculateCrcImagesCommand.parse(message);
    }

    /**
     * Tests exceptions of the function createData(): no data
     */
    @Test
    public final void testCreateDataExceptionNoData() {
        thrown.expect(IllegalArgumentException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data received "
                + "for CalculateCrcImages command when creating TDP message"));

        // incorrect message (no data)
        CalculateCrcImagesCommand.createData(new ArrayList<Integer>());
    }

}
