/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class SetBrightnessCommand.
 */
public class TdpCommandSetBrightnessTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse() {
        TdpBrightnessTable table = new TdpBrightnessTable();

        try {
            final ByteBuffer data = TdpUtils.createBuffer(11);
            data.put((byte)0x00); // 0%
            data.put((byte)0x0B); // 11%
            data.put((byte)0x16); // 22%
            data.put((byte)0x21); // 33%
            data.put((byte)0x2C); // 44%
            data.put((byte)0x37); // 55%
            data.put((byte)0x42); // 66%
            data.put((byte)0x4D); // 77%
            data.put((byte)0x58); // 88%
            data.put((byte)0x63); // 99%
            data.put((byte)0x64); // 100%

            // correct message (no timeout)
            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_BRIGHTNESS);
            message.addAddress(1);
            message.setData(data);
            table = SetBrightnessCommand.parse(message);
            int value = 0;
            for (int i = 0; i < 10; i++) {
                Assert.assertEquals(value, table.getBrightnessTable().get(i).intValue());
                value += 11;
            }
            Assert.assertEquals(100, table.getBrightnessTable().get(10).intValue());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong command id
     */
    @Test
    public final void testParseException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in command"));

        // incorrect message (wrong command id)
        final ByteBuffer data = TdpUtils.createBuffer(11);
        data.put((byte)0x00); // 0%

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in command"));

        // incorrect message (response)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.SET_BRIGHTNESS);
        message.addAddress(1);
        message.setData(data);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Incorrect percentage value
     */
    @Test
    public final void testParseException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Incorrect value"));

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(11);
        data.put((byte)0x00); // 0%
        data.put((byte)0x0B); // 11%
        data.put((byte)0x16); // 22%
        data.put((byte)0x21); // 33%
        data.put((byte)0x2C); // 44%
        data.put((byte)0x37); // 55%
        data.put((byte)0x42); // 66%
        data.put((byte)0x4D); // 77%
        data.put((byte)0x58); // 88%
        data.put((byte)0x63); // 99%
        data.put((byte)0x65); // 101%

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_BRIGHTNESS);
        message.addAddress(1);
        message.setData(data);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Too few entries
     */
    @Test
    public final void testParseException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("The brightness table must contain exactly"));

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(10);
        data.put((byte)0x00); // 0%
        data.put((byte)0x0B); // 11%
        data.put((byte)0x16); // 22%
        data.put((byte)0x21); // 33%
        data.put((byte)0x2C); // 44%
        data.put((byte)0x37); // 55%
        data.put((byte)0x42); // 66%
        data.put((byte)0x4D); // 77%
        data.put((byte)0x58); // 88%
        data.put((byte)0x63); // 99%

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_BRIGHTNESS);
        message.addAddress(1);
        message.setData(data);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Too much data
     */
    @Test
    public final void testParseException5() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("The brightness table must contain exactly"));

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(12);
        data.put((byte)0x00); // 0%
        data.put((byte)0x0B); // 11%
        data.put((byte)0x16); // 22%
        data.put((byte)0x21); // 33%
        data.put((byte)0x2C); // 44%
        data.put((byte)0x37); // 55%
        data.put((byte)0x42); // 66%
        data.put((byte)0x4D); // 77%
        data.put((byte)0x58); // 88%
        data.put((byte)0x63); // 99%
        data.put((byte)0x64); // 100%
        data.put((byte)0x64); // 100%

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_BRIGHTNESS);
        message.addAddress(1);
        message.setData(data);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseException6() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_BRIGHTNESS);
        message.addAddress(1);
        message.setData(null);
        SetBrightnessCommand.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpBrightnessTable brightnessTable = new TdpBrightnessTable();
        List<Integer> table = new ArrayList<>();
        table.add(0);
        table.add(10);
        table.add(20);
        table.add(30);
        table.add(40);
        table.add(50);
        table.add(60);
        table.add(70);
        table.add(80);
        table.add(90);
        table.add(100);
        brightnessTable.setBrightnessTable(table);
        SetBrightnessCommand.createData(brightnessTable);
    }

}
