/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpExternalLighting;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpLightingState;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class SetExternalCommand.
 */
public class TdpCommandSetLightingTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse() {
        TdpExternalLighting settings = new TdpExternalLighting();
        final ETdpLightingState state = ETdpLightingState.AUTOMATIC;

        try {
            final ByteBuffer data = TdpUtils.createBuffer(1);
            data.put((byte)state.getValue());

            // correct message (no timeout)
            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
            message.addAddress(1);
            message.setData(data);
            settings = SetExternalLightingCommand.parse(message);
            Assert.assertEquals(state, settings.getState());

        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong command id
     */
    @Test
    public final void testParseException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in command"));

        // incorrect message (wrong command id)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x00);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in command"));

        // incorrect message (response)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
        message.addAddress(1);
        message.setData(data);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Unknown timeout mode
     */
    @Test
    public final void testParseException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown value for state"));

        // incorrect message (unknown timeout mode)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x04);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
        message.addAddress(1);
        message.setData(data);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
        message.addAddress(1);
        message.setData(null);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): too much data
     */
    @Test
    public final void testParseException6() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        // incorrect message (mode = clear image, but image slot is also present)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x01);
        data.put((byte)0x01);
        data.put((byte)0x87);
        data.put((byte)0x68);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
        message.addAddress(1);
        message.setData(data);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): too much data
     */
    @Test
    public final void testParseException7() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        // incorrect message (mode = no timeout, but timeout data is present)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x00);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_EXTERNAL_LIGHTING);
        message.addAddress(1);
        message.setData(data);
        SetExternalLightingCommand.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpExternalLighting lighting = new TdpExternalLighting();
        lighting.setState(ETdpLightingState.AUTOMATIC);
        SetExternalLightingCommand.createData(lighting);
    }

}
