/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpTimeoutSettings;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpTimeoutMode;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class SetTimeoutCommand.
 */
public class TdpCommandSetTimeoutTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParseCorrectMessages() {
        TdpTimeoutSettings settings = new TdpTimeoutSettings();

        try {
            final ByteBuffer data = TdpUtils.createBuffer(1);
            data.put((byte)0x00);

            // correct message (no timeout)
            TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
            message.addAddress(1);
            message.setData(data);
            settings = SetTimeoutCommand.parse(message);
            Assert.assertEquals(ETdpTimeoutMode.NO_TIMEOUT, settings.getTimeoutMode());

            // correct message (clear display and timeout of 1000s)
            final ByteBuffer data1 = TdpUtils.createBuffer(3);
            data1.put((byte)0x01);
            data1.put((byte)0x87);
            data1.put((byte)0x68);

            message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
            message.addAddress(1);
            message.setData(data1);
            settings = SetTimeoutCommand.parse(message);
            Assert.assertEquals(ETdpTimeoutMode.CLEAR_DISPLAY, settings.getTimeoutMode());
            Assert.assertEquals(1000, settings.getTimeoutInSec());

            // correct message (show image, timeout of 1s, image slot 1000)
            final ByteBuffer data2 = TdpUtils.createBuffer(4);
            data2.put((byte)0x02);
            data2.put((byte)0x01);
            data2.put((byte)0x87);
            data2.put((byte)0x68);

            message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
            message.addAddress(1);
            message.setData(data2);
            settings = SetTimeoutCommand.parse(message);
            Assert.assertEquals(ETdpTimeoutMode.SHOW_IMAGE, settings.getTimeoutMode());
            Assert.assertEquals(1, settings.getTimeoutInSec());
            Assert.assertEquals(1000, settings.getImageSlot());

        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parse()
     */
    @Test
    public final void testCreateDataCorrectMessages() {
        final TdpTimeoutSettings settings = new TdpTimeoutSettings();
        ByteBuffer data;

        try {
            settings.setTimeoutMode(ETdpTimeoutMode.NO_TIMEOUT);
            data = SetTimeoutCommand.createData(settings);
            Assert.assertEquals(settings.getTimeoutMode().getValue(), data.get());

            // correct message (no timeout)
            settings.setTimeoutMode(ETdpTimeoutMode.CLEAR_DISPLAY);
            settings.setTimeoutInSec(128);
            data = SetTimeoutCommand.createData(settings);
            Assert.assertEquals(settings.getTimeoutMode().getValue(), data.get());
            Assert.assertEquals(settings.getTimeoutInSec(), VlqUtils.parseVlqData(data));

            // correct message (clear display and timeout of 1000s)
            settings.setTimeoutMode(ETdpTimeoutMode.SHOW_IMAGE);
            settings.setTimeoutInSec(128);
            settings.setImageSlot(16383);
            data = SetTimeoutCommand.createData(settings);
            Assert.assertEquals(settings.getTimeoutMode().getValue(), data.get());
            Assert.assertEquals(settings.getTimeoutInSec(), VlqUtils.parseVlqData(data));
            Assert.assertEquals(settings.getImageSlot(), VlqUtils.parseVlqData(data));
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong command id
     */
    @Test
    public final void testParseExceptionWrongCommandId() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in command"));

        // incorrect message (wrong command id)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x04);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseExceptionWrongMessageType() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in command"));

        // incorrect message (response)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        message.addAddress(1);
        message.setData(data);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Unknown timeout mode
     */
    @Test
    public final void testParseExceptionUnknownTimeoutValue() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown value for timeout mode"));

        // incorrect message (unknown timeout mode)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x04);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        message.addAddress(1);
        message.setData(data);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseExceptionNoDataNull() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        message.addAddress(1);
        message.setData(null);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): too much data
     */
    @Test
    public final void testParseExceptionTooMuchData1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        // incorrect message (mode = clear image, but image slot is also present)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x01);
        data.put((byte)0x01);
        data.put((byte)0x87);
        data.put((byte)0x68);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        message.addAddress(1);
        message.setData(data);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): too much data
     */
    @Test
    public final void testParseExceptionTooMuchData2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        // incorrect message (mode = no timeout, but timeout data is present)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x00);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        message.addAddress(1);
        message.setData(data);
        SetTimeoutCommand.parse(message);
    }

    /**
     * Tests the function createData(): no timeout value when mode = CLEAR_DISPLAY
     */
    @Test
    public final void testCreateDataExceptionNoTimeoutValue() {
        thrown.expect(IllegalArgumentException.class);
        thrown.expectMessage(CoreMatchers.containsString("Timeout value is mandatory when timeout mode "
                + "not is NO_TIMEOUT in SetCommunicationTimeout command"));

        final TdpTimeoutSettings settings = new TdpTimeoutSettings();

        // correct message (no timeout)
        settings.setTimeoutMode(ETdpTimeoutMode.CLEAR_DISPLAY);
        SetTimeoutCommand.createData(settings);
    }

    /**
     * Tests the function createData(): no image slot when mode = SHOW_IMAGE
     */
    @Test
    public final void testCreateDataExceptionNoImageslot() {
        thrown.expect(IllegalArgumentException.class);
        thrown.expectMessage(CoreMatchers.containsString("No negative value expected for VLQ quantity"));

        final TdpTimeoutSettings settings = new TdpTimeoutSettings();

        // correct message (no timeout)
        settings.setTimeoutMode(ETdpTimeoutMode.SHOW_IMAGE);
        settings.setTimeoutInSec(50);
        SetTimeoutCommand.createData(settings);
    }
}
