/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpDisplayType;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class TdpResponseDisplayProperties.
 */
public class TdpResponseDisplayPropertiesTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseDataNONMATRIX() {
        TdpDisplayProperties properties = new TdpDisplayProperties();
        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.ROTATIONPANEL;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        // correct message (only mandatory fields)
        try {
            final ByteBuffer data = TdpUtils.createBuffer(90);
            data.put((byte)0x40); // protocol version tag
            data.put((byte)protocolVersion);
            data.put((byte)0x41); // display type tag
            data.put((byte)displayType.getValue());
            data.put((byte)0xC2); // supplier and product type tag
            data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
            data.put((byte)0xC3); // serial number tag
            data.put(VlqUtils.getVlqBytes(serialnumber.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
            data.put((byte)0xC4); // software version tag
            data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
            message.addAddress(1);
            message.setData(data);

            properties = DisplayPropertiesResponse.parse(message);

            Assert.assertEquals(protocolVersion, properties.getProtocolVersion());
            Assert.assertEquals(displayType, properties.getDisplayType());
            Assert.assertEquals(supplierAndProductType, properties.getSupplierAndProductType());
            Assert.assertEquals(serialnumber, properties.getSerialNumber());
            Assert.assertEquals(softwareVersion, properties.getSoftwareVersion());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseDataMATRIX() {
        TdpDisplayProperties properties = new TdpDisplayProperties();
        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 127;
        final int displayWidth = 128;
        final int nrFixedImages = 16383;
        final int nrWritableImages = 0;
        final int nrSlideShowImages = 2;
        final ByteBuffer rgbColors = ByteBuffer.allocate(3);
        rgbColors.put((byte)0x80);
        rgbColors.put((byte)0xFF);
        rgbColors.put((byte)0x01);
        rgbColors.rewind();

        // correct message (all fields)
        try {
            final ByteBuffer data = TdpUtils.createBuffer(109);
            data.put((byte)0x40); // protocol version tag
            data.put((byte)protocolVersion);
            data.put((byte)0x41); // display type tag
            data.put((byte)displayType.getValue());
            data.put((byte)0xC2); // supplier and product type tag
            data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
            data.put((byte)0xC3); // serial number tag
            data.put(VlqUtils.getVlqBytes(serialnumber.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
            data.put((byte)0xC4); // software version tag
            data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
            data.put((byte)0x05); // external lighting tag
            data.put((byte)0x50); // display height tag
            data.put(VlqUtils.getVlqBytes(displayHeight));
            data.put((byte)0x91); // display width tag
            data.put(VlqUtils.getVlqBytes(displayWidth));
            data.put((byte)0x92); // nr. fixed images tag
            data.put(VlqUtils.getVlqBytes(nrFixedImages));
            data.put((byte)0x53); // nr. writable images tag
            data.put(VlqUtils.getVlqBytes(nrWritableImages));
            data.put((byte)0x54); // slide show tag
            data.put((byte)nrSlideShowImages);
            data.put((byte)0xD5); // rgb colors tag
            data.put(VlqUtils.getVlqBytes(3));
            data.put(rgbColors);
            data.put((byte)0x17); // png supported tag

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
            message.addAddress(1);
            message.setData(data);

            properties = DisplayPropertiesResponse.parse(message);

            Assert.assertEquals(protocolVersion, properties.getProtocolVersion());
            Assert.assertEquals(displayType, properties.getDisplayType());
            Assert.assertEquals(supplierAndProductType, properties.getSupplierAndProductType());
            Assert.assertEquals(serialnumber, properties.getSerialNumber());
            Assert.assertEquals(softwareVersion, properties.getSoftwareVersion());
            Assert.assertEquals(true, properties.hasExternalLighting());
            Assert.assertEquals(displayHeight, properties.getDisplayHeightInPixels());
            Assert.assertEquals(displayWidth, properties.getDisplayWidthInPixels());
            Assert.assertEquals(nrFixedImages, properties.getNrFixedImages());
            Assert.assertEquals(nrWritableImages, properties.getNrWritableImages());
            Assert.assertEquals(true, properties.hasSlideShow());
            Assert.assertEquals(nrSlideShowImages, properties.getNrSlideShowImages());
            Assert.assertEquals(rgbColors.rewind(), properties.getRgbColors());
            Assert.assertEquals(true, properties.isPngSupported());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parseData(): invalid tag
     */
    @Test
    public final void testParseDataException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown tag in data of DisplayProperties message"));

        // incorrect message (first tag id is invalid)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x4F);
        data.put((byte)0x01);
        data.put((byte)0x41);
        data.put((byte)0x03);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid message type
     */
    @Test
    public final void testParseDataException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type"));

        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x40);
        data.put((byte)0x01);
        data.put((byte)0x41);
        data.put((byte)0x03);

        // incorrect message (message type is invalid)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid number of addresses
     */
    @Test
    public final void testParseDataException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong number of addresses"));

        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x40);
        data.put((byte)0x01);
        data.put((byte)0x41);
        data.put((byte)0x03);

        // incorrect message (number of addresses is invalid)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.addAddress(2);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong command id
     */
    @Test
    public final void testParseDataException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id"));

        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x40);
        data.put((byte)0x01);
        data.put((byte)0x41);
        data.put((byte)0x03);

        // incorrect message (command is invalid)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.addAddress(2);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid tag
     */
    @Test
    public final void testParseDataException5() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown tag"));

        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)0x4F);
        data.put((byte)0x01);
        data.put((byte)0x41);
        data.put((byte)0x03);

        // incorrect message (command is invalid)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Protocol version tag missing
     */
    @Test
    public final void testParseDataException6() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag ProtocolVersion missing"));

        final ETdpDisplayType displayType = ETdpDisplayType.ROTATIONPANEL;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(88);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Display type tag missing
     */
    @Test
    public final void testParseDataException7() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag DisplayType missing"));

        final int protocolVersion = 1;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(88);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Supplier and product type tag missing
     */
    @Test
    public final void testParseDataException8() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag SupplierAndProductType missing"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.ROTATIONPANEL;
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(58);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Serial number tag missing
     */
    @Test
    public final void testParseDataException9() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag SerialNumber missing"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.ROTATIONPANEL;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String softwareVersion = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(78);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Software version tag missing
     */
    @Test
    public final void testParseDataException10() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag SoftwareVersion missing"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.ROTATIONPANEL;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(78);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): No display height when display type is MATRIX
     */
    @Test
    public final void testParseDataException11() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Display properties must contain data for "
                + "display height when display type = MATRIX"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        final ByteBuffer data = TdpUtils.createBuffer(90);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): No display width when display type is MATRIX
     */
    @Test
    public final void testParseDataException12() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Display properties must contain data for display width "
                + "when display type = MATRIX"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 10;

        final ByteBuffer data = TdpUtils.createBuffer(92);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
        data.put((byte)0x50); // display height tag
        data.put((byte)displayHeight);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): No RGBColors/FixedColorPallete when display type is MATRIX
     */
    @Test
    public final void testParseDataException13() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Display properties must contain data for RGB_COLORS "
                + "or FIXED_COLOR_PALETTE when display type = MATRIX, neither is present"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 10;
        final int displayWidth = 10;

        final ByteBuffer data = TdpUtils.createBuffer(94);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
        data.put((byte)0x50); // display height tag
        data.put((byte)displayHeight);
        data.put((byte)0x51); // display width tag
        data.put((byte)displayWidth);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Both RGBColors and FixedColorPallete when display type is MATRIX
     */
    @Test
    public final void testParseDataException14() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Display properties must contain data for RGB_COLORS "
                + "or FIXED_COLOR_PALETTE when display type = MATRIX, both are present"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 10;
        final int displayWidth = 10;
        final ByteBuffer rgbColors = ByteBuffer.allocate(3);
        rgbColors.put((byte)0x80);
        rgbColors.put((byte)0xFF);
        rgbColors.put((byte)0x01);
        rgbColors.rewind();
        final ByteBuffer fixedColorsPalette = ByteBuffer.allocate(3);
        fixedColorsPalette.put((byte)0x80);
        fixedColorsPalette.put((byte)0xFF);
        fixedColorsPalette.put((byte)0x01);
        fixedColorsPalette.rewind();

        final ByteBuffer data = TdpUtils.createBuffer(104);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
        data.put((byte)0x50); // display height tag
        data.put((byte)displayHeight);
        data.put((byte)0x51); // display width tag
        data.put((byte)displayWidth);
        data.put((byte)0xD5); // rgb colors tag
        data.put(VlqUtils.getVlqBytes(3));
        data.put(rgbColors);
        data.put((byte)0xD6); // fixed color palette tag
        data.put(VlqUtils.getVlqBytes(3));
        data.put(fixedColorsPalette);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): FixedColorPallete data is not a multiple of 3 bytes
     */
    @Test
    public final void testParseDataException15() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong data length received"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 10;
        final int displayWidth = 10;
        final ByteBuffer fixedColorsPalette = ByteBuffer.allocate(4);
        fixedColorsPalette.put((byte)0x80);
        fixedColorsPalette.put((byte)0xFF);
        fixedColorsPalette.put((byte)0x01);
        fixedColorsPalette.put((byte)0x00);
        fixedColorsPalette.rewind();

        final ByteBuffer data = TdpUtils.createBuffer(104);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
        data.put((byte)0x50); // display height tag
        data.put((byte)displayHeight);
        data.put((byte)0x51); // display width tag
        data.put((byte)displayWidth);
        data.put((byte)0xD6); // fixed color palette tag
        data.put(VlqUtils.getVlqBytes(4));
        data.put(fixedColorsPalette);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): Unknown value for display type
     */
    @Test
    public final void testParseDataException16() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown value for DisplayType"));

        final int protocolVersion = 1;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";

        // correct message (only mandatory fields)
        final ByteBuffer data = TdpUtils.createBuffer(90);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)0xFF);
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): no data
     */
    @Test
    public final void testParseDataException17() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function parseData(): too much data
     */
    @Test
    public final void testParseDataException19() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        final int protocolVersion = 1;
        final ETdpDisplayType displayType = ETdpDisplayType.MATRIX;
        final String supplierAndProductType = "1234567890123456789012345678901234567890";
        final String serialnumber = "12345678901234567890";
        final String softwareVersion = "12345678901234567890";
        final int displayHeight = 127;
        final int displayWidth = 128;
        final int nrFixedImages = 16383;
        final int nrWritableImages = 0;
        final int nrSlideShowImages = 2;
        final ByteBuffer rgbColors = ByteBuffer.allocate(3);
        rgbColors.put((byte)0x80);
        rgbColors.put((byte)0xFF);
        rgbColors.put((byte)0x01);
        rgbColors.rewind();

        final ByteBuffer data = TdpUtils.createBuffer(110);
        data.put((byte)0x40); // protocol version tag
        data.put((byte)protocolVersion);
        data.put((byte)0x41); // display type tag
        data.put((byte)displayType.getValue());
        data.put((byte)0xC2); // supplier and product type tag
        data.put(VlqUtils.getVlqBytes(supplierAndProductType.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(supplierAndProductType));
        data.put((byte)0xC3); // serial number tag
        data.put(VlqUtils.getVlqBytes(serialnumber.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(serialnumber));
        data.put((byte)0xC4); // software version tag
        data.put(VlqUtils.getVlqBytes(softwareVersion.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(softwareVersion));
        data.put((byte)0x05); // external lighting tag
        data.put((byte)0x50); // display height tag
        data.put((byte)displayHeight);
        data.put((byte)0x91); // display width tag
        data.putShort((short)displayWidth);
        data.put((byte)0x92); // nr fixed images tag
        data.putShort((short)nrFixedImages);
        data.put((byte)0x53); // nr writable images tag
        data.put((byte)nrWritableImages);
        data.put((byte)0x54); // slide show tag
        data.put((byte)nrSlideShowImages);
        data.put((byte)0xD5); // rgb colors tag
        data.put(VlqUtils.getVlqBytes(3));
        data.put(rgbColors);
        data.put((byte)0x17); // png supported tag
        data.put((byte)0x00); // extra byte

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        DisplayPropertiesResponse.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpDisplayProperties displayProperties = new TdpDisplayProperties();
        
        displayProperties.setDisplayAddress(VlqUtils.MAX_VALUE);
        displayProperties.setProtocolVersion(255);
        displayProperties.setDisplayType(ETdpDisplayType.MATRIX);
        displayProperties.setSupplierAndProductType("Product display, Type a.b.c");
        displayProperties.setSerialNumber("#asdfghjkl");
        displayProperties.setSoftwareVersion("version a.b.c");
        displayProperties.setExternalLighting(true);
        displayProperties.setDisplayHeightInPixels(VlqUtils.MAX_VALUE);
        displayProperties.setDisplayHeightInPixels(VlqUtils.MAX_VALUE);
        displayProperties.setNrFixedImages(VlqUtils.MAX_VALUE);
        displayProperties.setNrWritableImages(VlqUtils.MAX_VALUE);
        displayProperties.setSlideShow(true);
        displayProperties.setRgbColors(ByteBufferUtils.stringToByteBuffer("ABC"));
        displayProperties.setPngSupported(true);
        DisplayPropertiesResponse.createData(displayProperties);
        
        // do some extra tests on TdpDisplayProperties
        displayProperties.setNrFixedImages(10);
        Assert.assertTrue(displayProperties.isFixedImage(5));
        Assert.assertFalse(displayProperties.isFixedImage(15));
        Assert.assertTrue(displayProperties.isWritableImage(15));
        Assert.assertFalse(displayProperties.isWritableImage(5));
    }
}
