/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpNotification;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpNotification;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class Notification response
 */
public class TdpResponseNotificationTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse() {
        List<TdpNotification> notifications = new ArrayList<TdpNotification>();

        try {
            final ByteBuffer data = TdpUtils.createBuffer(2);
            data.put((byte)0x41); // tag Communication error
            data.put((byte)ETdpCommunicationError.ERROR_CRC.getValue());

            // correct message (only one notification)
            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.NOTIFICATION);
            message.addAddress(1);
            message.setData(data);
            notifications = NotificationResponse.parse(message);
            Assert.assertEquals(1, notifications.size());
            Assert.assertEquals(ETdpNotification.COMMUNICATION_ERROR, notifications.get(0).getTag());
            Assert.assertEquals(ETdpCommunicationError.ERROR_CRC,
                    ((TdpCommunicationError)notifications.get(0)).getCommunicationError());

        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse2() {
        List<TdpNotification> notifications = new ArrayList<TdpNotification>();

        try {
            final ByteBuffer data = TdpUtils.createBuffer(16);
            data.put((byte)0x41); // tag Communication error
            data.put((byte)ETdpCommunicationError.ERROR_CRC.getValue());
            data.put((byte)ETdpNotification.SOFTWARE_FAILURE.getValue());
            data.put((byte)ETdpNotification.HARDWARE_FAILURE.getValue());
            data.put((byte)ETdpNotification.COLD_RESTART.getValue());
            data.put((byte)ETdpNotification.WARM_RESTART.getValue());
            data.put((byte)ETdpNotification.COMMUNICATION_TIMEOUT.getValue());
            data.put((byte)ETdpNotification.NON_CRITICAL_DISPLAY_DEFECT.getValue());
            data.put((byte)ETdpNotification.CRITICAL_DISPLAY_DEFECT.getValue());
            data.put((byte)ETdpNotification.INTRUSION.getValue());
            data.put((byte)ETdpNotification.EXTERNAL_LIGHTING_DEFECT.getValue());
            data.put((byte)ETdpNotification.HEATING_DEFECT.getValue());
            data.put((byte)ETdpNotification.COOLING_DEFECT.getValue());
            data.put((byte)ETdpNotification.TEMPERATURE_LOW.getValue());
            data.put((byte)ETdpNotification.TEMPERATURE_HIGH.getValue());
            data.put((byte)ETdpNotification.LUMINANCE_SENSOR_DEFECT.getValue());

            // correct message (all notifications)
            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.NOTIFICATION);
            message.addAddress(1);
            message.setData(data);
            notifications = NotificationResponse.parse(message);
            Assert.assertEquals(15, notifications.size());
            Assert.assertEquals(ETdpNotification.COMMUNICATION_ERROR, notifications.get(0).getTag());
            Assert.assertEquals(ETdpNotification.SOFTWARE_FAILURE, notifications.get(1).getTag());
            Assert.assertEquals(ETdpNotification.HARDWARE_FAILURE, notifications.get(2).getTag());
            Assert.assertEquals(ETdpNotification.COLD_RESTART, notifications.get(3).getTag());
            Assert.assertEquals(ETdpNotification.WARM_RESTART, notifications.get(4).getTag());
            Assert.assertEquals(ETdpNotification.COMMUNICATION_TIMEOUT, notifications.get(5).getTag());
            Assert.assertEquals(ETdpNotification.NON_CRITICAL_DISPLAY_DEFECT, notifications.get(6).getTag());
            Assert.assertEquals(ETdpNotification.CRITICAL_DISPLAY_DEFECT, notifications.get(7).getTag());
            Assert.assertEquals(ETdpNotification.INTRUSION, notifications.get(8).getTag());
            Assert.assertEquals(ETdpNotification.EXTERNAL_LIGHTING_DEFECT, notifications.get(9).getTag());
            Assert.assertEquals(ETdpNotification.HEATING_DEFECT, notifications.get(10).getTag());
            Assert.assertEquals(ETdpNotification.COOLING_DEFECT, notifications.get(11).getTag());
            Assert.assertEquals(ETdpNotification.TEMPERATURE_LOW, notifications.get(12).getTag());
            Assert.assertEquals(ETdpNotification.TEMPERATURE_HIGH, notifications.get(13).getTag());
            Assert.assertEquals(ETdpNotification.LUMINANCE_SENSOR_DEFECT, notifications.get(14).getTag());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong command id
     */
    @Test
    public final void testParseException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in response"));

        // incorrect message (wrong command id)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)ETdpNotification.COLD_RESTART.getValue());

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);
        NotificationResponse.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in response"));

        // incorrect message (response)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)ETdpNotification.COLD_RESTART.getValue());


        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.NOTIFICATION);
        message.addAddress(1);
        message.setData(data);
        NotificationResponse.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): unknown notification tag
     */
    @Test
    public final void testParseException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected tag in data of Notification response"));

        // incorrect message (response)
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x3F);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.NOTIFICATION);
        message.addAddress(1);
        message.setData(data);
        NotificationResponse.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): missing data for communication error
     */
    @Test
    public final void testParseException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Not all expected data received in message"));

        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x41); // tag Communication error

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.NOTIFICATION);
        message.addAddress(1);
        message.setData(data);
        NotificationResponse.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): unknown communication error
     */
    @Test
    public final void testParseException5() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value ("));
        // actual unexpected value
        thrown.expectMessage(CoreMatchers.containsString(") in data of communication error of Notification response"));

        final ByteBuffer data = TdpUtils.createBuffer(2);
        data.put((byte)0x41); // tag Communication error
        data.put((byte)0x04);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.NOTIFICATION);
        message.addAddress(1);
        message.setData(data);
        NotificationResponse.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseException6() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.NOTIFICATION);
        message.addAddress(1);
        message.setData(null);
        NotificationResponse.parse(message);
    }

    /**
     * Tests the function createData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        List<TdpNotification> notifications = new ArrayList<>();
        notifications.add(new TdpCommunicationError(ETdpCommunicationError.ERROR_UNKNOWN_COMMAND));

        NotificationResponse.createData(notifications);
    }

}
