/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.disperanto.data.TdpStatus;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class StatusResponse.
 */
public class TdpResponseStatusTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseData() {
        TdpStatus status = new TdpStatus();

        final int firstImageSlot = 127;
        final short firstImageCrc = 255;
        final int secondImageSlot = 16383;
        final short secondImageCrc = (short)65535;

        // correct message (only mandatory fields)
        try {
            final ByteBuffer data = TdpUtils.createBuffer(9);
            data.put((byte)0xC1); // shown image tag
            data.put(VlqUtils.getVlqBytes(7)); // length of shown images
            data.put(VlqUtils.getVlqBytes(firstImageSlot)); // image slot first image
            data.putShort(firstImageCrc); // crc first image
            data.put(VlqUtils.getVlqBytes(secondImageSlot)); // image slot second image
            data.putShort(secondImageCrc); // crc second image

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.STATUS);
            message.addAddress(1);
            message.setData(data);

            status = StatusResponse.parse(message);

            Assert.assertEquals(2, status.getShownImages().size());
            Assert.assertEquals(true, status.getShownImages().containsKey(firstImageSlot));
            Assert.assertEquals(firstImageCrc, status.getShownImages().get(firstImageSlot).shortValue());
            Assert.assertEquals(true, status.getShownImages().containsKey(secondImageSlot));
            Assert.assertEquals(secondImageCrc, status.getShownImages().get(secondImageSlot).shortValue());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseData2() {
        TdpStatus status = new TdpStatus();

        final int firstImageSlot = 127;
        final short firstImageCrc = 255;
        final int secondImageSlot = 16383;
        final short secondImageCrc = (short)65535;
        final int displayBrightness = 100;
        final int intensityExternalLighting = 0;
        final List<Integer> measuredLightIntensity = Arrays.asList(100, 50, 0);
        final String gps = "1234567890123456";
        final int temperature = 255;
        final int heating = 1;
        final int cooling = 1;

        // correct message (all fields)
        try {
            final ByteBuffer data = TdpUtils.createBuffer(42);
            data.put((byte)0xC1); // shown image tag
            data.put(VlqUtils.getVlqBytes(7)); // length of shown images
            data.put(VlqUtils.getVlqBytes(firstImageSlot)); // image slot first image
            data.putShort(firstImageCrc); // crc first image
            data.put(VlqUtils.getVlqBytes(secondImageSlot)); // image slot second image
            data.putShort(secondImageCrc); // crc second image
            data.put((byte)0x42); // brightness tag
            data.put((byte)displayBrightness);
            data.put((byte)0x43); // intensity external lighting tag
            data.put((byte)intensityExternalLighting);
            data.put((byte)0xC4); // measured light intensity tag
            data.put(VlqUtils.getVlqBytes(measuredLightIntensity.size()));
            final Iterator<Integer> it = measuredLightIntensity.iterator();
            while (it.hasNext()) {
                data.put((byte)it.next().intValue());
            }
            data.put((byte)0xC5); // gps tag
            data.put(VlqUtils.getVlqBytes(gps.length()));
            data.put(ByteBufferUtils.stringToByteBuffer(gps));
            data.put((byte)0x46); // temperature tag
            data.put((byte)temperature);
            data.put((byte)0x47); // heating tag
            data.put((byte)heating);
            data.put((byte)0x48); // cooling tag
            data.put((byte)cooling);

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.RESPONSE);
            message.setCommand(ETdpCommand.STATUS);
            message.addAddress(1);
            message.setData(data);

            status = StatusResponse.parse(message);

            Assert.assertEquals(2, status.getShownImages().size());
            Assert.assertEquals(true, status.getShownImages().containsKey(firstImageSlot));
            Assert.assertEquals(firstImageCrc, status.getShownImages().get(firstImageSlot).shortValue());
            Assert.assertEquals(true, status.getShownImages().containsKey(secondImageSlot));
            Assert.assertEquals(secondImageCrc, status.getShownImages().get(secondImageSlot).shortValue());
            Assert.assertEquals(displayBrightness, status.getBrightnessInPercentage().intValue());
            Assert.assertEquals(intensityExternalLighting, status.getIntensityExternalLighting().intValue());
            Assert.assertEquals(measuredLightIntensity, status.getMeasuredLightIntensity());
            Assert.assertEquals(gps, status.getGpsPosition());
            Assert.assertEquals(temperature, status.getTemperature().intValue() & 0xFF);
            Assert.assertEquals(heating, status.getHeating().intValue());
            Assert.assertEquals(cooling, status.getCooling().intValue());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parseData(): invalid tag
     */
    @Test
    public final void testParseDataException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown tag in data of Status message"));

        final ByteBuffer data = TdpUtils.createBuffer(2);
        data.put((byte)0x4F);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid message type
     */
    @Test
    public final void testParseDataException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid number of addresses
     */
    @Test
    public final void testParseDataException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong number of addresses"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.addAddress(2);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong command id
     */
    @Test
    public final void testParseDataException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid tag
     */
    @Test
    public final void testParseDataException5() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown tag"));

        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x0F);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): mandatory tag missing
     */
    @Test
    public final void testParseDataException6() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Mandatory tag ShownImages missing"));

        final ByteBuffer data = TdpUtils.createBuffer(2);
        data.put((byte)0x42); // brightness tag
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong value for brightness
     */
    @Test
    public final void testParseDataException7() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value for Brightness data"));

        final ByteBuffer data = TdpUtils.createBuffer(9);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(3)); // length of shown images
        data.put(VlqUtils.getVlqBytes(1)); // image slot first image
        data.putShort((short)1); // crc first image
        data.put((byte)0x42); // brightness tag
        data.put((byte)101);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong value for intensity external lighting
     */
    @Test
    public final void testParseDataException8() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value for IntensityExternalLighting data"));

        final ByteBuffer data = TdpUtils.createBuffer(11);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(3)); // length of shown images
        data.put(VlqUtils.getVlqBytes(1)); // image slot first image
        data.putShort((short)1); // crc first image
        data.put((byte)0x43); // intensity external lighting
        data.put((byte)101);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong value for measured light intensity
     */
    @Test
    public final void testParseDataException9() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value for MeasuredLightIntensity data"));

        final ByteBuffer data = TdpUtils.createBuffer(16);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(3)); // length of shown images
        data.put(VlqUtils.getVlqBytes(1)); // image slot first image
        data.putShort((short)1); // crc first image
        data.put((byte)0xC4); // measured light intensity tag
        data.put(VlqUtils.getVlqBytes(3));
        data.put((byte)0);
        data.put((byte)100);
        data.put((byte)101);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong value for heating
     */
    @Test
    public final void testParseDataException10() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value for Heating data"));

        final ByteBuffer data = TdpUtils.createBuffer(9);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(3)); // length of shown images
        data.put(VlqUtils.getVlqBytes(1)); // image slot first image
        data.putShort((short)1); // crc first image
        data.put((byte)0x47); // heating tag
        data.put((byte)2);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong value for cooling
     */
    @Test
    public final void testParseDataException11() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unexpected value for Cooling data"));

        final ByteBuffer data = TdpUtils.createBuffer(9);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(3)); // length of shown images
        data.put(VlqUtils.getVlqBytes(1)); // image slot first image
        data.putShort((short)1); // crc first image
        data.put((byte)0x48); // cooling tag
        data.put((byte)2);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): no data
     */
    @Test
    public final void testParseDataException12() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function parseData(): too much data
     */
    @Test
    public final void testParseDataException14() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Too much data received for message"));

        final int firstImageSlot = 127;
        final short firstImageCrc = 255;
        final int secondImageSlot = 16383;
        final short secondImageCrc = (short)65535;
        final int displayBrightness = 100;
        final int intensityExternalLighting = 0;
        final List<Integer> measuredLightIntensity = Arrays.asList(100, 50, 0);
        final String gps = "1234567890123456";
        final int temperature = 255;
        final int heating = 1;
        final int cooling = 1;

        // correct message (all fields)
        final ByteBuffer data = TdpUtils.createBuffer(43);
        data.put((byte)0xC1); // shown image tag
        data.put(VlqUtils.getVlqBytes(7)); // length of shown images
        data.put(VlqUtils.getVlqBytes(firstImageSlot)); // image slot first image
        data.putShort(firstImageCrc); // crc first image
        data.put(VlqUtils.getVlqBytes(secondImageSlot)); // image slot second image
        data.putShort(secondImageCrc); // crc second image
        data.put((byte)0x42); // brightness tag
        data.put((byte)displayBrightness);
        data.put((byte)0x43); // intensity external lighting tag
        data.put((byte)intensityExternalLighting);
        data.put((byte)0xC4); // measured light intensity tag
        data.put(VlqUtils.getVlqBytes(measuredLightIntensity.size()));
        final Iterator<Integer> it = measuredLightIntensity.iterator();
        while (it.hasNext()) {
            data.put((byte)it.next().intValue());
        }
        data.put((byte)0xC5); // gps tag
        data.put(VlqUtils.getVlqBytes(gps.length()));
        data.put(ByteBufferUtils.stringToByteBuffer(gps));
        data.put((byte)0x46); // temperature tag
        data.put((byte)temperature);
        data.put((byte)0x47); // heating tag
        data.put((byte)heating);
        data.put((byte)0x48); // cooling tag
        data.put((byte)cooling);
        data.put((byte)0x00);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.STATUS);
        message.addAddress(1);
        message.setData(data);

        StatusResponse.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpStatus status = new TdpStatus();

        Map<Integer, Short> shownImages = new HashMap<>();
        shownImages.put(VlqUtils.MAX_VALUE, Short.MAX_VALUE);
        status.setShownImages(shownImages);
        status.setBrightnessInPercentage(100);
        status.setIntensityExternalLighting(100);
        List<Integer> measuredLightIntensity = new ArrayList<>();
        measuredLightIntensity.add(100);
        status.setMeasuredLightIntensity(measuredLightIntensity);
        status.setGpsPosition("1234567890123456");
        status.setTemperature(255);
        status.setHeating(1);
        status.setCooling(1);
        StatusResponse.createData(status);
    }
}
