/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
   list of conditions and the following disclaimer in the documentation and/or
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 */
package org.disperanto.protocol;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.List;

import org.disperanto.communication.UdpReceiver;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpDisplayType;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.DisplayPropertiesResponse;
import org.disperanto.messages.TdpMessage;
import org.disperanto.utils.TdpUtils;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

/**
 * JUnit tests for the class TdpProtocol.
 */
public class TdpProtocolTest {
    private TdpProtocol tdpProtocol;
    private UdpReceiver connection;

    /**
     * Default setup for each test.
     */
    @Before
    public final void setUp() throws IOException {
        connection = new UdpReceiver(3456, 1000);
        connection.start();
        tdpProtocol = new TdpProtocolExchanger(connection);
    }

    @After
    public final void cleanUp() {
        connection.stop();
    }

    /**
     * Tests the function calculateMessageLength()
     */
    @Test
    public final void testCalculateMessageLength() {
        // Message with no data and one address
        TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1),
                null);
        int length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(7, length);

        // Message with one data byte and two addresses
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1, 2),
                ByteBuffer.allocate(1));
        length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(9, length);

        // Message with two data bytes and three addresses
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1, 2, 3),
                ByteBuffer.allocate(2));
        length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(11, length);

        // Message with 127 data bytes and four addresses
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1, 2, 3, 4),
                ByteBuffer.allocate(127));
        length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(137, length);

        // Message with 128 data bytes and four addresses
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1, 2, 3, 4),
                ByteBuffer.allocate(128));
        length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(139, length);

        // Message with 16384 data bytes and one address
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1),
                ByteBuffer.allocate(16384));
        length = tdpProtocol.calculateMessageLength(message);
        Assert.assertEquals(16393, length);
    }

    /**
     * Tests the function createStartByte()
     */
    @Test
    public final void testCreateStartByte() {
        // DisplayProperties command
        TdpMessage message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES,
                Arrays.asList(1, 2), null);

        int startByte = tdpProtocol.createStartByte(message, true);
        // expected result is 1100 0010
        Assert.assertEquals(0xC2, startByte);

        startByte = tdpProtocol.createStartByte(message, false);
        // expected result is 1000 0010
        Assert.assertEquals(0x82, startByte);

        // DisplayProperties response
        message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1), null);

        startByte = tdpProtocol.createStartByte(message, true);
        // expected result is 0100 0001
        Assert.assertEquals(0x41, startByte);

        startByte = tdpProtocol.createStartByte(message, false);
        // expected result is 0000 0001
        Assert.assertEquals(0x01, startByte);
    }

    /**
     * Tests the function parseMessageType()
     */
    @Test
    public final void testParseMessageType() throws TdpProtocolException {
        Assert.assertEquals(ETdpMessageType.RESPONSE, tdpProtocol.parseMessageType((byte)0x01));
        Assert.assertEquals(ETdpMessageType.COMMAND, tdpProtocol.parseMessageType((byte)0xC2));
    }

    /**
     * Tests the function parseLastMessageFlag()
     */
    @Test
    public final void testParseLastMessageFlag() {
        boolean lastMessage = false;

        // last message
        lastMessage = tdpProtocol.parseLastMessageFlag((byte)0xC2);
        Assert.assertEquals(true, lastMessage);

        // not last message
        lastMessage = tdpProtocol.parseLastMessageFlag((byte)0x01);
        Assert.assertEquals(false, lastMessage);
    }

    /**
     * Tests the function parseNrAddresses()
     */
    @Test
    public final void testParseNrAddresses() {
        int nrAddresses = 0;

        // 1 address
        nrAddresses = tdpProtocol.parseNrAddresses((byte)0xC1);
        Assert.assertEquals(1, nrAddresses);

        // 63 addresses
        nrAddresses = tdpProtocol.parseNrAddresses((byte)0xFF);
        Assert.assertEquals(63, nrAddresses);
    }

    /**
     * Tests the function parseCommand()
     */
    @Test
    public final void testParseCommand() {
        Assert.assertEquals(ETdpCommand.DISPLAY_PROPERTIES, tdpProtocol.parseCommand((byte)0x01));
        Assert.assertEquals(ETdpCommand.STATUS, tdpProtocol.parseCommand((byte)0x02));
    }

    /**
     * Tests the function getAddressBuffer()
     */
    @Test
    public final void testGetAddressBuffer() {
        List<Integer> addresses = Arrays.asList(1);
        ByteBuffer addressBuffer = tdpProtocol.getAddressBuffer(addresses);
        addressBuffer.rewind();
        Assert.assertEquals((byte)0x01, addressBuffer.get());

        addresses = Arrays.asList(1, 2);
        addressBuffer = tdpProtocol.getAddressBuffer(addresses);
        addressBuffer.rewind();
        Assert.assertEquals((byte)0x01, addressBuffer.get());
        Assert.assertEquals((byte)0x02, addressBuffer.get());
    }

    /**
     * Tests the function createMessageData()
     */
    @Test
    public final void testCreateMessageData() {
        TdpMessage message = null;
        ByteBuffer data = null;

        // Command with no data
        message = new TdpMessage(ETdpMessageType.COMMAND, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(21, 255), null);
        data = tdpProtocol.createMessageData(message, true);

        data.rewind();
        Assert.assertEquals(8, data.capacity());
        // start byte: bit 7: 1 (ETdpMessageType.COMMAND), bit 6 = 1 (last message), bit 0-5 = 2 (two addresses)
        Assert.assertEquals((byte)0xC2, data.get());
        // message number is not check, value is unknown
        data.get();
        // address = 21
        Assert.assertEquals((byte)0x15, data.get());
        // address = 255
        Assert.assertEquals((byte)0xFF, data.get());
        // command byte: bit 7 = 0, bit 6 = 0 (no data), bit 0-5 = 1 (ETdpCommand.DISPLAY_PROPERTIES)
        Assert.assertEquals((byte)0x01, data.get());
        // data length byte: 0
        Assert.assertEquals((byte)0x00, data.get());
        // ignore CRC

        final TdpDisplayProperties displayProperties = new TdpDisplayProperties();
        displayProperties.setProtocolVersion(2);
        displayProperties.setDisplayType(ETdpDisplayType.ROTATIONPANEL);
        displayProperties.setSupplierAndProductType("000");
        displayProperties.setSerialNumber("12");
        displayProperties.setSoftwareVersion("1");
        final ByteBuffer rgbColors = ByteBuffer.allocate(3);
        rgbColors.put((byte)0x01);
        rgbColors.put((byte)0x02);
        rgbColors.put((byte)0x03);
        displayProperties.setRgbColors(rgbColors);

        message = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(43),
                DisplayPropertiesResponse.createData(displayProperties));

        data = tdpProtocol.createMessageData(message, false);

        data.rewind();
        Assert.assertEquals(26, data.capacity());
        // start byte: bit 7: 0 (ETdpMessageType.RESPONSE), bit 6 = 0 (not last message), bit 0-5 = 1 (one address)
        Assert.assertEquals((byte)0x01, data.get());
        // message number is not check, value is unknown
        data.get();
        // address = 43
        Assert.assertEquals((byte)0x2B, data.get());
        // command byte: bit 7 = 1, bit 6 = 1 (vlq data length), bit 0-5 = 1 (ETdpCommand.DISPLAY_PROPERTIES)
        Assert.assertEquals((byte)0x01, data.get());
        // vlq byte: bit 7 = 0 (no more vlq-bytes) bit 0-6 = 19 (data length)
        Assert.assertEquals((byte)0x13, data.get());
        // data bytes
        Assert.assertEquals((byte)0x40, data.get()); // protocol version
        Assert.assertEquals((byte)0x02, data.get());
        Assert.assertEquals((byte)0x41, data.get()); // display type
        Assert.assertEquals((byte)0x05, data.get()); // ROTATIONPANEL
        Assert.assertEquals((byte)0xC2, data.get()); // supplier and product type
        Assert.assertEquals((byte)0x03, data.get()); // vlq
        Assert.assertEquals((byte)0x30, data.get());
        Assert.assertEquals((byte)0x30, data.get());
        Assert.assertEquals((byte)0x30, data.get());
        Assert.assertEquals((byte)0x83, data.get()); // serial number
        Assert.assertEquals((byte)0x31, data.get());
        Assert.assertEquals((byte)0x32, data.get());
        Assert.assertEquals((byte)0x44, data.get()); // software version
        Assert.assertEquals((byte)0x31, data.get());
        Assert.assertEquals((byte)0xD5, data.get()); // rgb colors
        Assert.assertEquals((byte)0x03, data.get()); // vlq
        Assert.assertEquals((byte)0x01, data.get());
        Assert.assertEquals((byte)0x02, data.get());
        Assert.assertEquals((byte)0x03, data.get());
        // ignore CRC
    }

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseData() {
        List<TdpMessage> messages = null;

        // Tests 2 commands, other commands are similar.

        // Create data buffer for two commands
        final ByteBuffer data = ByteBuffer.allocate(15);

        // Display Properties command
        ByteBuffer message = ByteBuffer.allocate(8);
        // start byte: bit 7: 1 (ETdpMessageType.COMMAND), bit 6 = 0 (not last message), bit 0-5 = 2 (two addresses)
        message.put((byte)0x82);
        // message number = 123
        message.put((byte)0x7B);
        // address = 1
        message.put((byte)0x01);
        // address = 255
        message.put((byte)0xFF);
        // command byte: ETdpCommand.DISPLAY_PROPERTIES
        message.put((byte)0x01);
        // data length: 0
        message.put((byte)0x00);
        // CRC bytes
        message.putShort(TdpUtils.calculateCrc(message.array()));
        message.rewind();
        data.put(message);

        // Status command
        message = ByteBuffer.allocate(7);
        // start byte: bit 7: 1 (ETdpMessageType.COMMAND), bit 6 = 1 (last message), bit 0-5 = 1 (one address)
        message.put((byte)0xC1);
        // message number = 124
        message.put((byte)0x7C);
        // address = 10
        message.put((byte)0x0A);
        // command byte: ETdpCommand.STATUS
        message.put((byte)0x02);
        // data length: 0
        message.put((byte)0x00);
        // CRC bytes
        message.putShort(TdpUtils.calculateCrc(message.array()));
        message.rewind();
        data.put(message);

        try {
            messages = tdpProtocol.parseData(new ByteArrayInputStream(data.array()));
        } catch (Exception e) {
            e.printStackTrace();
            Assert.fail();
            return;
        }

        Assert.assertEquals(2, messages.size());
        // check Display Properties command
        Assert.assertEquals(ETdpMessageType.COMMAND, messages.get(0).getMessageType());
        Assert.assertEquals(123, messages.get(0).getMessageNumber());
        Assert.assertEquals((byte)0x01, messages.get(0).getAddresses().get(0).byteValue());
        Assert.assertEquals((byte)0xFF, messages.get(0).getAddresses().get(1).byteValue());
        Assert.assertEquals(ETdpCommand.DISPLAY_PROPERTIES, messages.get(0).getCommand());
        Assert.assertEquals(0, messages.get(0).getData().capacity());
        // check Status command
        Assert.assertEquals(ETdpMessageType.COMMAND, messages.get(1).getMessageType());
        Assert.assertEquals(124, messages.get(1).getMessageNumber());
        Assert.assertEquals((byte)0x0A, messages.get(1).getAddresses().get(0).byteValue());
        Assert.assertEquals(ETdpCommand.STATUS, messages.get(1).getCommand());
        Assert.assertEquals(0, messages.get(1).getData().capacity());
    }

    /**
     * Tests the function readMessageFromBuffer()
     */
    @Test
    public final void testReadMessageFromBuffer() {
        TdpMessage original = null;
        TdpMessage result = null;
        boolean lastMessage = false;

        try {
            original = new TdpMessage(ETdpMessageType.RESPONSE, ETdpCommand.DISPLAY_PROPERTIES, Arrays.asList(1), null);
            result = new TdpMessage();
            lastMessage = tdpProtocol.parseMessageData(new ByteArrayInputStream(
                    tdpProtocol.createMessageData(original, true).array()), result);
        } catch (Exception e) {
            Assert.fail();
            return;
        }
        Assert.assertEquals(true, lastMessage);
        Assert.assertEquals(original.getMessageType(), result.getMessageType());
        Assert.assertEquals(original.getCommand(), result.getCommand());
        Assert.assertEquals(original.getAddresses(), result.getAddresses());
        Assert.assertEquals(0, result.getData().capacity());
    }

    @Test
    public void parseUnknownCommand() throws IOException, TdpProtocolException {
        // Sequence of two messages: invalid command, status
        try (InputStream in = getClass().getResourceAsStream("invalid-command-id.bin")) {
            List<TdpMessage> messages = tdpProtocol.parseData(in);

            Assert.assertEquals(2, messages.size());

            // Unknown command is parsed
            TdpMessage message = messages.get(0);
            Assert.assertEquals(ETdpMessageType.COMMAND, message.getMessageType());
            Assert.assertEquals(ETdpCommand.UNKNOWN, message.getCommand());

            // Encountering an unknown command doesn't prevent the following message from being parsed
            message = messages.get(1);
            Assert.assertEquals(ETdpMessageType.COMMAND, message.getMessageType());
            Assert.assertEquals(ETdpCommand.STATUS, message.getCommand());
        }
    }
}
