/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.refapp.displaycontroller;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.disperanto.communication.UdpReceiver;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpNotification;
import org.disperanto.data.TdpStatus;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.TdpCommandFactory;
import org.disperanto.messages.TdpMessage;
import org.disperanto.protocol.ITdpReceiveListener;
import org.disperanto.protocol.TdpProtocolReceiver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Handles requests for displays using Disperanto.
 */
public class TdpDisplayControllerService implements ITdpReceiveListener {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpDisplayControllerService.class);

    private TdpProtocolReceiver receiver;
    private UdpReceiver connection;
    private HashMap<Integer, Display> displayMap = new HashMap<Integer, Display>();

    /**
     * Constructor
     * 
     * @param displays list of displays that are available for this display controller
     * @param receivePort Port number used to receive packets
     * @param timeoutMs Timeout for UDP communication in milliseconds
     */
    public TdpDisplayControllerService(List<Display> displays, int receivePort, int timeoutMs) {
        LOGGER.info("Display Controller created, timeout: " + timeoutMs);

        // Set displays in a map that has the address as key value
        if (displays != null) {
            for (Display display : displays) {
                displayMap.put(display.getAddress(), display);
            }
        }

        // Setup UDP connection
        connection = new UdpReceiver(receivePort, timeoutMs);

        // Initialize receiver
        receiver = new TdpProtocolReceiver(connection);
        receiver.addReceiveListener(this);

        // start receiving
        final ExecutorService threadExecutor = Executors.newFixedThreadPool(1);
        threadExecutor.execute(receiver);
    }

    /**
     * Start the receiver thread.
     * @throws IOException when connection cannot be started.
     */
    public final void start() throws IOException {
        connection.start();
    }

    /**
     * Stops the receiver thread that listens to incoming messages and stops the connection.
     */
    public final void stop() {
        try {
            receiver.stop();
        } catch (InterruptedException e) {
            LOGGER.error("Error trying to stop", e);
        }
        connection.stop();
    }

    @Override
    public final void commandReceived(List<TdpMessage> commands) throws TdpProtocolException, IOException {
        final List<TdpMessage> responseList = new ArrayList<TdpMessage>();

        // Parse commands
        for (TdpMessage command : commands) {
            if (command.getMessageType() != ETdpMessageType.COMMAND) {
                throw new TdpProtocolException("Wrong message type in command");
            }

            for (Integer address : command.getAddresses()) {
                switch (command.getCommand()) {
                case DISPLAY_PROPERTIES:
                    responseList.add(createDisplayPropertiesResponse(address.intValue(), command.getMessageNumber()));
                    break;
                case STATUS:
                    responseList.add(createStatusResponse(address.intValue(), command.getMessageNumber()));
                    break;
                default:
                    throw new TdpProtocolException(
                            "Unexpected command value received while requesting display properties");
                }
            }

            for (Integer address : command.getAddresses()) {
                responseList.add(createNotifications(address.intValue()));
            }
        }

        receiver.send(responseList);

    }

    /**
     * Adds an display to the display controller.
     * 
     * @param display Display that is available for the display controller
     */
    public final void addDisplay(Display display) {
        displayMap.put(display.getAddress(), display);
    }

    /**
     * removes an display of the display controller.
     * 
     * @param display Display that is available for the display controller
     */
    public final void removeDisplay(Display display) {
        displayMap.remove(display.getAddress());
    }

    /**
     * Creates for each addressed display an response with display properties.
     * 
     * @param messageNumber
     * 
     * @param addresses Display addresses the command is sent to
     * @return TDP response for a display properties command
     * @throws IOException
     */
    private TdpMessage createDisplayPropertiesResponse(Integer address, int messageNumber) throws IOException {
        final TdpDisplayProperties displayProperties = displayMap.get(address).getDisplayProperties();
        return TdpCommandFactory.createDisplayPropertiesResponse(address, messageNumber, displayProperties);
    }

    /**
     * Creates for each addressed display an response with the display status.
     * 
     * @param addresses Display addresses the command is sent to
     * @return TDP response for a status command
     * @throws IOException
     */
    private TdpMessage createStatusResponse(Integer address, int messageNumber) throws IOException {
        final TdpStatus status = displayMap.get(address).getStatus();
        return TdpCommandFactory.createStatusResponse(address, messageNumber, status);
    }

    private TdpMessage createNotifications(int address) {
        final List<TdpNotification> notificationList = displayMap.get(address).getNotifications();

        return TdpCommandFactory.createNotification(address, notificationList);
    }
}
